#!/usr/bin/env python
# encoding: utf-8
"""
iasi_cloud_extract.py

requires EUGENE toolkit for python, numpy, scipy (for f2py)

Created by Keoni on 2008-03-10.
Copyright (c) 2008 University of Wisconsin SSEC. All rights reserved.
"""

import sys, os, re
from numpy import array,zeros
import eugene
import uuid
from datetime import datetime
from collections import defaultdict
from subprocess import check_call as sh
from subprocess import CalledProcessError
from glob import glob
from shutil import rmtree
# see http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/81549
import SocketServer
import SimpleXMLRPCServer
import xmlrpclib 
import cStringIO

TMP = os.environ.get('TMP','/tmp')

from metahoard.iasi.avhrr_bt import avhrr_extract, openclose, writeout  # from f2py -c avhrr_bt.f90 -m avhrr_bt --fcompiler=gfortran

import logging
LOG = logging.getLogger( "iasi_cloud_extract" )

RE_IASI_DATE=re.compile('IASI_xxx_1C_M02_([0-9]{8})([0-9]{6})Z')
FMT_CLUSTERS = 'clusters_metop02_%8s_%6s_%05d.txt'   # ymd, hms, orbit
IASIMASK_BAT = '/home/rayg/aapp/MAIA/iasimask.bat'
RSYNC_ANCIL = 'sands::incoming/MAIA/nwp' # % yyyymmdd string
RSYNC_ANCIL_TMP = 'sands::tmp/MAIA/nwp' # % yyyymmdd string

def clusters_extract(infilename):
    """extract data from IASI native format files and write out a text file 
    match the format of the output of rd_iasi_1c.f90
    """
    prod = eugene.Product(infilename)
    hdr = prod.get('MPHR')
    numorb = hdr.ORBIT_START
    if numorb==0: 
        raise ValueError('bad orbit number')
    stem = os.path.split(infilename)[1]
    (yyyymmdd,hhmmss) = RE_IASI_DATE.match(stem).groups()
    iouta = openclose(0,stem,numorb)
    (cluster_filename,) = glob(FMT_CLUSTERS % (yyyymmdd,hhmmss,numorb))
    record_names = prod._records_['mdr-1c']
    for record_name in record_names:
        line = int(*re.findall( '\[([0-9]+)\]', record_name))
        GEPSDatIasi = array(prod.get('%s.GEPSDatIasi[]' % (record_name)))
        GGeoSondAnglesSUN = array(prod.get('%s.GGeoSondAnglesSUN[][][]' % (record_name)))
        GGeoSondLoc = array(prod.get('%s.GGeoSondLoc[][][]' % (record_name)))
        GGeoSondAnglesSUN = array(prod.get('%s.GGeoSondAnglesSUN[][][]' % (record_name)))
        GGeoSondAnglesMETOP = array(prod.get('%s.GGeoSondAnglesMETOP[][][]' % (record_name)))
        GCcsRadAnalNbClass = array(prod.get('%s.GCcsRadAnalNbClass[][]' % (record_name)))
        GCcsRadAnalWgt = array(prod.get('%s.GCcsRadAnalWgt[][]' % (record_name)))
        GCcsRadAnalMean = array(prod.get('%s.GCcsRadAnalMean[][]' % (record_name)))
        GCcsRadAnalStd = array(prod.get('%s.GCcsRadAnalStd[][]' % (record_name)))
        for jpos in range(30):        
            etime = GEPSDatIasi[jpos]  # this is a datetime object
            # millisecond of the day
            itime = int(etime.microsecond/1000) + etime.hour*3600000 + etime.minute*60000 + etime.second*1000
            for ifov in range(4):
                satlon = GGeoSondLoc[jpos,ifov,0]
                satlat = GGeoSondLoc[jpos,ifov,1]
                solzen = GGeoSondAnglesSUN[jpos,ifov,0]
                solazm = GGeoSondAnglesSUN[jpos,ifov,1]
                satzen = GGeoSondAnglesMETOP[jpos,ifov,0]
                satazm = GGeoSondAnglesMETOP[jpos,ifov,1]
            
                nbclass = GCcsRadAnalNbClass[jpos,ifov] 
                # convert this to the reverse polish notation equivalent for fortran
                # .squeeze removes all 1-dimensions: 1x1x7x6 -> 7x6
                # .transpose flips from 7x6 to 6x7
                # then copy into a fortran-contiguous array with float32 type
                wgt = array(GCcsRadAnalWgt[jpos,ifov].squeeze().transpose(),dtype='f',order='FORTRAN')
                gcmean = array(GCcsRadAnalMean[jpos,ifov].squeeze().transpose(),dtype='f',order='FORTRAN')
                gcstd = array(GCcsRadAnalStd[jpos,ifov].squeeze().transpose(),dtype='f',order='FORTRAN') 
                
                # call fortran function
                frac, tab_tb, tab_sd = avhrr_extract( wgt, gcmean, gcstd )
                writeout(iouta, line+1,jpos+1,ifov+1,itime,nbclass, frac,tab_tb,tab_sd, satlat,satlon,satzen,satazm,solzen,solazm)
    # close the file
    openclose(iouta,'',0)
    return cluster_filename, yyyymmdd, hhmmss, numorb                               
        
class maia_record(object):
    """ MAIA output record as parsed from text line
    """
    RE_FLOAT = re.compile(r'[.\-0-9NnAaEe]+')  # cover exponential and NaN
    LIGHT_TABLE = { None: '?', 0: 'n', 1: 't', 2: 'd', 3: 'g' } # night, twilight, day, glint
    SURFACE_TABLE = { None: '?', 0: 'S', 1: 'L', 2: 'C' }  # sea, land, coast
    
    def _str_type(self):
        return '%c%c' % (self.day_time, self.surface_type)
    
    def __repr__(self):
        return '<maia_record at scan %(scan_line)d FOR %(field_of_regard)d FOV %(field_of_view)d>' % vars(self)
    
    def __init__(self, line, yyyymmdd_str = None):
        values = self.RE_FLOAT.findall(line)
        # normalize to zero-based values used by EUGENE and elsewhere in iasi_cache
        # NOTE: see iasi_cache table declarations; the naming convention is kept common to simplify merging
        (   self.scan_line,  
            self.field_index, # 0..29
            self.detector_index ) = [int(float(x))-1 for x in values[0:3]]
        
        (   self.satellite_latitude,
            self.satellite_longitude,
            self.satellite_zenith_angle ) = [float(x) for x in values[3:6]]
        
        self.field_of_regard_time = int(float(values[6]))
        
        self.day_time = int(float(values[7]))
        self.surface_type = int(float(values[8]))
            
        (   self.clear_percent,
            self.cloudy_percent,
            self.layer_count ) = [float(x) for x in values[9:12]]
            
        (   self.surface_temp,
            self.twometer_temp,
            self.surface_pressure ) = [float(x) for x in values[47:50]]
            
        if yyyymmdd_str:
            y,m,d = int(yyyymmdd_str[:4]), int(yyyymmdd_str[4:6]), int(yyyymmdd_str[6:8])
            US = (self.field_of_regard_time % 1000) * 1000
            S = int(self.field_of_regard_time / 1000) 
            M = int(S / 60) % 60
            H = int(S/3600) % 60
            S %= 60            
            self.when = datetime( y,m,d, H,M,S, US )
        else:
            self.when = None

TEST_MAIA_LINE = "            22.00   26.00    3.00   66.29  -10.45   39.64    78293310.0     0.00     0.00     0.00   100.00     4.00    16.00    60.18   -16.88     1.85     0.00    16.00    21.32   -23.46     2.09     0.00     8.00    17.70   -12.88     0.53     0.00    17.00     0.80   -27.99     0.25     0.00     0.00     0.00     0.00     0.00     0.00     0.00     0.00     0.00     0.00     0.00     0.00     0.00     0.00     0.00     0.00   274.90   276.27   993.20"

RE_CLUSTERS = re.compile(r'.*?clusters_metop(?P<sat>\d\d)_(?P<yyyymmdd>\d{8})_(?P<hhmmss>\d{6})_(?P<orbno>\d+).*')
RE_MAIA = re.compile(r'.*?iasimask_metop(?P<sat>\d\d)_(?P<yyyymmdd>\d{8})_(?P<hhmmss>\d{6})_(?P<orbno>\d+).*')

def read_maia_file(filename, fp = None):
    """generate maia records for a file of a given name, or a file-like object
    optional file pointer, for the case of wanting to delete the file prior to using the generator
    """
    nfo = RE_MAIA.match(filename).groupdict()
    if not fp: 
        LOG.debug('opening %s' % filename)
        fp = file(filename,'rt')
    for line in fp: 
        yield maia_record(line,yyyymmdd_str=nfo['yyyymmdd'])

def read_maia_file_as_dict(filename):
    """read a maia file, returning a dict of (scan,field,detector) -> dictionary
    """
    zult = defaultdict(list)
    for rec in read_maia_file(filename):
        zult[(rec.scan_line, rec.field_index, rec.detector_index)] = vars(rec)
    return zult

def _ymd(yyyymmdd_s):
    ymd = datetime.strptime( yyyymmdd_s, '%Y%m%d' )
    return ymd

def _daysago(yyyymmdd_s):
    "return the number of days ago a given YYYYMMDD is, in GMT"
    ymd = _ymd(yyyymmdd_s)
    now = datetime.utcnow()
    return (now-ymd).days

# table of which satellite numbers in clusters filenames correspond to which input to iasimask
SATS = { '02': 'metopa' }

def cache_nwp(target_dir, yyyymmdd_s, base = RSYNC_ANCIL, tmpbase = RSYNC_ANCIL_TMP):
    """recover ancillary files needed for a given day from the SAN
    untar and decompress them to the target directory
    build softlink directories similar to construction of /home/halw/nwp/ 
    return paths of the files
    """
    ymd = _ymd(yyyymmdd_s)
    link_dir = os.path.join(target_dir, ymd.strftime('%Y/%m%d%y'))
    mmddyy_s = ymd.strftime('%m%d%y')
    
    if not os.path.isdir(link_dir):
        os.makedirs(link_dir)
        
    # do we already have this data locally?
    def list_of_nwp_files(suffix=''):
        return list(glob(os.path.join(target_dir,'maia_ancillary_%s??_??%s' % (yyyymmdd_s,suffix))))
        
    previously = list_of_nwp_files()
    if len(previously)==20: 
        LOG.debug('ancillary data already available for this date')
        return previously

    # check TMP for uncompressed files first
    LOG.debug( 'downloading ancillary data from cache' )
    tmp_url = '%s/%s/maia_ancillary_%s??_??' % (tmpbase,mmddyy_s,yyyymmdd_s)
    try:
        sh( ['rsync', '-ua', tmp_url, target_dir] )
    except CalledProcessError:
        LOG.debug('nothing found in temporary cache; error from rsync (ignored)')
        pass

    # did we catch what we wanted?
    fromcache = list_of_nwp_files()
    LOG.debug('from cache: %s' % str(fromcache))
    
    if len(fromcache)==0: # then we go for the tar file in the long-term storage
        url = '%s/anc_%s.tar' % (base, yyyymmdd_s)
        tar_filename = url.split('/')[-1]
        tar_pathname = os.path.join(target_dir,tar_filename)

        LOG.debug( 'downloading ancillary data from storage' )
        sh( ['rsync', '-ua', url, tar_pathname] )

        LOG.debug( 'unpacking ancillary data' )
        sh( ['tar', 'xf', tar_pathname, '-C', target_dir] )

        LOG.debug( 'decompressing ancillary data' )
        nwpgzfiles = list_of_nwp_files('.gz')
        assert(len(nwpgzfiles)==20)
        sh(['gunzip', '-f'] + nwpgzfiles)
        os.unlink(tar_pathname)
    
    # link them to the link_dir
    lonf = list_of_nwp_files()
    for pn in lonf:
        fn = os.path.split(pn)[-1]
        linkname = os.path.join(link_dir,fn)
        if not os.path.exists(linkname):
            os.symlink( os.path.join('../..',fn), linkname )
    
    return lonf

def run_maia(filename, nwp_dir = None):
    """ run maia on a given clusters file
    return pathname of output
    output will land in the same directory as the input file
    optional directory path to NWP data (see cache_nwp)
    """
    oldpwd = os.getcwd()
    dirn = os.path.split(os.path.abspath(filename))[0]
    os.chdir(dirn)
    nfo = RE_CLUSTERS.match(filename).groupdict()
    os.environ['PATH']+=':.'
    os.environ['idir']=dirn
    os.environ['odir']=dirn
    wrk_prd = os.tempnam()
    os.makedirs(wrk_prd)
    os.environ['ARCHIPEL_dir_wrk_prd'] = wrk_prd # in MAIADIRS
    if nwp_dir is not None:
        cache_nwp(nwp_dir, nfo['yyyymmdd'])
        os.environ['apath']=nwp_dir
    # run maia for data from today on metop-a
    sh([IASIMASK_BAT, nfo['yyyymmdd'], SATS[nfo['sat']]])
    (fn,) = [fn for fn in glob('iasimask*%(sat)s*%(yyyymmdd)s*%(hhmmss)s*%(orbno)s*' % nfo) if RE_MAIA.match(fn)]
    os.chdir(oldpwd)
    # sh(['ls', '-l', wrk_prd])
    rmtree(wrk_prd)
    return os.path.join(dirn,fn)
    
def mask_extract(local_pathname, text_data=None, clusters_func = None, mask_func=None, keep=False, nwp_dir=TMP):
    """ use iasi_mask_extract to build cluster file if text_data is not provided
    __call__( iasi_native_pathname )
    __call__( clusters_filename, clusters_file_content )
    
    run maia
    read result
    return dictionary of field-of-view information per read_maia_file_as_dict
    clear out temporary files
    clusters_func and mask_func are optional callable data hooks which accept file-like objects
    """
    # extract the cluster file to the cwd if no data is provided
    workspace = None
    if not text_data:
        LOG.debug('extracting cluster data from native file %s' % local_pathname)
        clusters_pathname,_,_,_ = clusters_extract( local_pathname )
        LOG.debug('created clusters file %s' % clusters_pathname)
        if clusters_func: 
            clusters_func(file(clusters_pathname,'rt'))
    else:
        # write a temporary clusters file from the text data since maia needs that 
        clusters_pathname = os.path.split(local_pathname)[-1]
        workspace = os.path.join(TMP, str(uuid.uuid1()))
        LOG.debug('workspace is %s' % workspace)
        os.makedirs(workspace)
        clusters_pathname=os.path.join(workspace,clusters_pathname)
        LOG.debug('writing %s with clusters data...' % clusters_pathname)
        file(clusters_pathname,'wt').write(text_data)
        
    LOG.debug('running maia on %s with nwp data cache at %s' % (clusters_pathname, nwp_dir))
    output_filename = run_maia(clusters_pathname, nwp_dir)
    LOG.debug('iasimask file is %s' % output_filename)
    if mask_func: 
        mask_func(file(output_filename,'rt'))
    nfo = read_maia_file(output_filename, file(output_filename,'rt'))
    if workspace is not None and not keep:
        LOG.debug('flushing workspace %s' % workspace)
        rmtree(workspace)
    if keep:
        LOG.info('output mask file is %s' % output_filename)
    return nfo

class remote_mask_extract(object):
    """cloud extraction using a remote server
    usage: 
    rcx = remote_mask_extract( (host,port) )
    for fn in clusters_filenames:
        for maia_rec in rcx(fn):
            # use maia_record data
    """
    def __init__(self,(host,port)):
        self.server = xmlrpclib.Server('http://%s:%d' % (host,port))
        self.workspace = os.path.join(TMP, str(uuid.uuid1()))
        os.makedirs(self.workspace)
        LOG.debug('workspace is %s' % self.workspace)
    
    def __del__(self):
        rmtree(self.workspace)
    
    def __call__(self, iasi_pathname, text_data = None, clusters_func = None, mask_func=None, keep=False):
        """ 
        __call__( iasi_native_pathname )
        __call__( clusters_filename, clusters_file_content )
        identical function as mask_extract() but with RPC server inbetween
        
        extract clusters from a native format file if only a filename is provided
        send cluster file to a remote MAIA service
        collect output and return list of maia_records
        if text_data is provided, iasi_pathname is the name of the clusters file and text_data is its content
        clusters_func and mask_func are optional callable data hooks which accept file-like objects
        clusters_func is only called if the clusters file had to be generated
        """
        if text_data is not None:
            clusters_filename = os.path.split(iasi_pathname)[-1]
        else:
            iasi_pathname = os.path.abspath(iasi_pathname)
            
        oldcwd = os.getcwd()
        os.chdir(self.workspace)
        
        if text_data is None: # then extract clusters from raw file
            clusters_filename,_,_,_ = clusters_extract(iasi_pathname)
            LOG.debug('clusters file is %s' % clusters_filename)
            text_data = file(clusters_filename,'rt').read()
            if clusters_func: 
                clusters_func(file(clusters_filename,'rt'))
            os.unlink(clusters_filename)
                
        self.input = text_data
        name, mask_text = self.server.mask_extract(clusters_filename, text_data)        
        
        nfo = RE_MAIA.match(name).groupdict()
        os.chdir(oldcwd)
        self.output = mask_text
        
        if mask_func:
            mask_func(cStringIO.StringIO(mask_text))
            
        fp = cStringIO.StringIO(mask_text)
        for line in fp: 
            yield maia_record(line,yyyymmdd_str=nfo['yyyymmdd'])

def rpc_mask_extract(clusters_filename, clusters_text):
    """ rpc front end; take in text, return text
    """
    workspace = os.path.join(TMP, str(uuid.uuid1()))
    LOG.debug("MAIA workspace is %s" % workspace)
    os.makedirs(workspace)
    os.chdir(workspace)
    file(clusters_filename,'wt').write(clusters_text)
    output_filename = run_maia(clusters_filename, nwp_dir=TMP)
    output_text = file(output_filename,'rt').read()
    os.chdir('/tmp')
    rmtree(workspace)
    return output_filename, output_text


class ForkingXMLRPCServer (SocketServer.ForkingMixIn,
                           SimpleXMLRPCServer.SimpleXMLRPCServer):
    pass

def server_main(port=7867):
    """process which serves what it's being fed
    for now, just send to stdout rather than serving it over a socket in XML format
    """
    fracas = ForkingXMLRPCServer( ('',port) )
    fracas.register_function(rpc_mask_extract, 'mask_extract')
    fracas.serve_forever()

def main():
    import optparse
    usage = """$Id: iasi_cloud_extract.py,v 1.31 2008/08/14 19:35:32 rayg Exp $
%prog [options] inputfile1 inputfile2 inputfile3 ...

Input files are IASI native format and must have full filenames
e.g. IASI_xxx_1C_M02_20080209170855Z_20080209171159Z_N_O_20080209185150Z__20080209185625
Output files will be of a form e.g. 
clusters_metop02_20080209_170855_06791.txt
"""
    parser = optparse.OptionParser(usage)
    parser.add_option('-t', '--test', dest="self_test",
                    action="store_true", default=False, help="run self-tests")            
    parser.add_option('-v', '--verbose', dest="verbose",
                    action="store_true", default=False, help="enable debug output")
    parser.add_option('-k', '--keep', dest="keep",
                    action="store_true", default=False, help="keep intermediate output")
    parser.add_option('-S', '--serve-maia', dest="server_port", help="serve MAIA on a TCP port")
    
    parser.add_option('-M', '--maia', dest='maia_host', help='host:port of maia server to use')
                                     
    # parser.add_option("-o", "--output", dest="output_filename",
    #                   help="output filename")
    # parser.add_option("-c", "--cdl", dest="cdl_filename",
    #                 help="CDL format template filename")
    # parser.add_option("-d", "--donor", dest="donor_filename",
    #                 help="OPTIONAL: existing netcdf retrieval file to act as pressure level donor")
    # parser.add_option("-q", "--quiet",
    #               action="store_false", dest="verbose", default=True,
    #               help="don't print status messages to stdout")
    (options, args) = parser.parse_args()
    if options.self_test:
        import doctest
        doctest.testmod()
        sys.exit(2)
    LOG.setLevel(logging.DEBUG if options.verbose else logging.INFO)

    if options.server_port:
        LOG.info('serving!')
        server_main(int(options.server_port))
        return 0

    if options.maia_host:
        LOG.info('running mask extraction...')
        if options.maia_host in ['localhost', '-']:
            maskmaker = mask_extract
        else:
            host,port = options.maia_host.split(':')
            port = int(port)
            LOG.info('consulting MAIA server at (%s,%d)' % (host,port))
            maskmaker = remote_mask_extract( (host,port) )
            LOG.debug('connected')
        from cPickle import dump
        for iasi_file in args:
            if re.match('clusters.*?\.txt',os.path.split(iasi_file)[-1]):
                # read the clusters in and don't worry about trying to read product files
                LOG.info('processing cluster file %s...' % iasi_file)
                mm = maskmaker(iasi_file, file(iasi_file,'rt').read(), keep=options.keep)
            else:
                # generate the clusters file as well as the mask
                LOG.info('processing native file %s...' % iasi_file)
                mm = maskmaker(iasi_file, keep=options.keep)
            data = [vars(rec) for rec in mm]
            LOG.debug('%d records loaded' % len(data))
            out_filename = 'mask_%s.pickle' % os.path.split(iasi_file)[1]
            LOG.info('writing %s...' % out_filename)
            dump(data, file(out_filename,'wb'))
        return 0

    for infilename in args:
        print 'processing %s...' % infilename
        clusters_extract( infilename )

    parser.error( 'incorrect arguments, try -h or --help.' )
    return 9

if __name__=='__main__':
    logging.basicConfig()
    sys.exit(main())
