#!/usr/bin/env python
# encoding: utf-8
"""
iasi_cache.py

Created by Keoni on 2007-11-29.
Copyright (c) 2007 University of Wisconsin SSEC. All rights reserved.

Setting up a cache:
Create database
initialize database
create IASI_cache directory on SAN
get upstream directory set up

ssh -L5432:bora2:5432 ash
python -m iasi_cache --iasi-db=postgres://rayg:c@nnab1s@localhost/iasi_cache 

requires miscellaneous routines from keoni 0.3.4 or better (see larch.ssec.wisc.edu)

"""

_CVS_IDENT = "$Id: iasi_cache.py,v 1.124 2008/08/18 21:38:21 rayg Exp $"

import sys
import os
import logging
import uuid
import re
import time
from cPickle import dump
from ftplib import FTP
from subprocess import check_call as sh
from datetime import datetime, timedelta
from time import gmtime, strftime
from collections import defaultdict

from sqlalchemy import *    
import sqlalchemy.orm as orm

from metahoard.iasi import tools as iasi
from metahoard.iasi.config import *
from metahoard.iasi.schema import *
from metahoard.iasi.transfer import *
from metahoard.iasi.util import *


LOG = logging.getLogger( __name__ )


def export(db, target_dir, granule_name, linefields, keep = False):
    """for each scan of interest in an IASI native file
    provide individual MAT files to store as files of the form iasi_YYYYMMDD_HHMMSS_sMM_fNN_dD.mat
    MM = scan number, NN is field number, D is detector number
    downloads granule if the name given is not a local filename
    """
    if not os.path.exists(target_dir): 
        LOG.info('creating %s.' % target_dir)
        os.makedirs(target_dir)
    granule_name, local_pathname, chosen_url, linenums = obtain_granule(db, target_dir, granule_name)
    linefields = expand_user_field_selection(linefields,linenums)
    LOG.debug('final field list: %s' % str(linefields))
    if not local_pathname:
        LOG.error('could not obtain %s' % granule_name)
        return
    emit_mat_files_from_search(target_dir, chosen_url or local_pathname, local_pathname, linefields)
    if not keep and chosen_url is not None:
        LOG.debug('removing %s' % local_pathname)
        os.unlink(local_pathname)
    else:
        print 'raw data file is at %s' % local_pathname

def merge_maia(db, granule_name, cloud_extract_func, force_reload = False, keep = False, fovs=True, dry_run=False):
    """
    downloads granule if the name given is not a local filename
    extracts cloud mask information using cloud_extract_func => { (scanline,field,detector): {data}, ... }
    merges it into the field-of-view table
    force_reload causes scans and fields for granule to be purged and re-loaded
    """
    assert(cloud_extract_func is not None)
    granule_name, local_pathname, chosen_url, linenums = obtain_granule(db, TMP, granule_name)
    if not local_pathname:
        LOG.error('could not obtain %s' % granule_name)
        return
    # extract cloud mask data
    try: 
        maia_info = cloud_extract_func(local_pathname)
    except ValueError:
        LOG.error("unable to process %s" % local_pathname)
        return
    uu = db.uuid_for_name(granule_name)
    if not force_reload:
        # sfd_indices = (scanline,field,detector)
        for sfd_indices in maia_info.keys():
            LOG.debug('updating %s' % str(sfd_indices))
            newdata = db.maia_columns(maia_info, *sfd_indices)
            LOG.debug("adding to %s: %s" % (str(sfd_indices),str(newdata)))
            if not dry_run:
                db.update_field(uu, sfd_indices, newdata)
    else: # forced reload of all scans and fields for the granule, half-purge effectively
        LOG.debug('deleting scanlines for %s' % uu)
        db.delete_scanlines(uu)
        # get the scanline metadata generator and set it to output individual fields of view, not fields of regard
        scanner = iasi.generate_scanline_metadata(iasi.open_product(local_pathname), fovs=fovs)
        db.insert_scanlines_from_generator(uu, scanner, maia_info, dry_run=dry_run) # fills the scan and    
    if not keep and chosen_url:
        LOG.debug('removing %s' % local_pathname)
        os.unlink(local_pathname)
    else:
        print 'raw data file is at %s' % local_pathname            
                    
def emit(db=None, target_dir='/tmp', box=None, startend=None, keywords=None, fields=None, keep = False, string_criteria=None):
    """for each scan of interest in an IASI native file
    provide individual MAT files to store as files of the form iasi_YYYYMMDD_HHMMSS_sMM_fNN_dD.mat
    MM = scan number, NN is field number, D is detector number
    """
    if not os.path.exists(target_dir): 
        LOG.info('creating %s.' % target_dir)
        os.makedirs(target_dir)
    for D in ui_query(db,box,startend,keywords,fields,string_criteria):
        name = D['name']
        urls = D['urls']
        keywords = D['keywords']
        fields = D['fields']
        # identify which url to try
        local_pathname = os.path.join(target_dir, name)
        urls = sort_urls_by_score(urls)
        LOG.debug('ordered download urls: %s' % urls)
        chosen_url = get_any_replica(urls,local_pathname)
        if not chosen_url:
            print '-'*72
            print name
            print 'Download of this granule failed!'
            print 'The fields of interest are:'
            print _format_search_results(fields)
        else:
            print 'downloaded '+chosen_url
            emit_mat_files_from_search(target_dir, chosen_url, local_pathname, fields)
            if not keep:
                os.unlink(local_pathname)
            else:
                print 'raw data file is at %s' % local_pathname

def _distill_file(local_pathname,name=None,keywords=None,fovs=True):
    """create distilled metadata dictionary and write it to a file relative to the current working directory
    returns the path of that data, inclusive of directories created
    that dictionary gets uploaded to storage and needs to be processed through MAIA before going into the database
    RAW -> distilled .p.gz -> SAN -> MAIA@zara -> database
    see _intake_distilled_file versus _intake_original_file
    """
    from iasi_cloud_extract import clusters_extract
    from gzip import GzipFile
    prod = iasi.open_product(local_pathname)
    meta = iasi.simple_metadata(prod)
    if name is None:
        name = os.path.split(local_pathname)[-1]
    meta['name'] = name
    # insert keywords, for now without value associated (FUTURE)
    kw = dict( (x,None) for x in meta.get('keywords',[]) )
    if keywords: kw.update(keywords)
    meta['keywords'] = kw
    # extract scan line metadata to a list we can plug back into the insertion function
    meta['scanlines'] = [x for x in iasi.generate_scanline_metadata(prod,fovs=fovs)]
    # big hunk of text for clusters file - this is used to generate cloud mask using MAIA
    clusters_filename, yyyymmdd, hhmmss, _ = clusters_extract(local_pathname)
    clusters_data = file(clusters_filename,'rb').read()
    os.unlink(clusters_filename)
    meta['clusters_data'] = clusters_data
    meta['clusters_filename'] = clusters_filename
    # manufacture a filename
    ymd = '%s/%s/%s' % (yyyymmdd[:4], yyyymmdd[4:6], yyyymmdd[6:8])
    # ymd = meta['start_time'].strftime('%Y/%m/%d')
    if not os.path.exists(ymd):
        os.makedirs(ymd)
    pathname = os.path.join(ymd,'meta_%s.p.gz' % name)
    # write a pickle which we can add cloud mask data to for later ingest
    dump(meta, GzipFile(pathname, 'wb'))
    return name, pathname

# THIS is retired in favor of fixing iasi_cloud_extract to allow clusters data to go in directly ---
#
# def _local_mask_extract(clusters_filename,clusters_data,dry_run=False):
#     # place the clusters data in a temporary directory with the original filename
#     TMP = os.tempnam()
#     os.makedirs(TMP)
#     clusters_pathname = os.path.join( TMP, os.path.split(clusters_filename)[-1] )
#     LOG.debug('writing %s' % clusters_pathname)
#     file(clusters_pathname,'wt').write(clusters_data)
#     
#     # run maia on the clusters file to get the mask file
#     from iasi_cloud_extract import run_maia, read_maia_file
#     LOG.debug('invoking MAIA locally')
#     iasimask_pathname = run_maia(clusters_pathname)
#     LOG.debug('reading mask file')
#     mask_info = read_maia_file(iasimask_pathname)
#     if not dry_run:
#         LOG.debug('uploading mask file %s' % iasimask_pathname)
#         _upload_mask_to_storage(os.path.split(iasimask_pathname)[-1], file(iasimask_pathname,'rt').read())
#     
#     from shutil import rmtree
#     LOG.debug('clearing out %s' % TMP)
#     rmtree(TMP)

def _intake_distilled_file(db,local_pathname,name=None,keywords=None,dry_run=False,cloud_extract=None,fovs=True):
    """given the .p.gz file of a previously distilled file (see _distill_file)
    perform a full database insert
    identical function to _intake_original_file, but works on distilled .p.gz files created with distill()
    this allows us to create distilled files on the PEATE and cloud mask on ZARA
    """
    from gzip import GzipFile
    from cPickle import load
    
    simple = load( GzipFile(local_pathname,'rb') )
    
    if name is None:
        name = simple['name']
        
    if cloud_extract:  # if a cloud extraction function is provided, use it to get cloud mask from cluster data
        maia_info = cloud_extract(simple['clusters_filename'], simple['clusters_data'])
        LOG.debug('%d records in MAIA table' % len(maia_info))
    else:
        maia_info = { }
    if cloud_extract and not fovs:
        LOG.warning("MAIA data comes in as fields of view! fovs option should be true")

    LOG.debug( 'simple metadata: %s' % simple)
    if not dry_run:
        myid = db.insert_simple_entry(**simple)
    else: # dummy id
        myid = uuid.uuid1()
    LOG.debug("id is %s" % myid)
    
    # insert keywords, for now without value associated (FUTURE)
    kw = simple['keywords']
    if keywords: kw.update(keywords)
    if not dry_run:
        db.insert_keywords(myid,kw)
    else:
        LOG.debug("keywords to be inserted: %s" % kw)
            
    # insert the predistilled scanlines as if the list was a generator
    db.insert_scanlines_from_generator(myid, simple['scanlines'], maia_info, dry_run=dry_run) # fills the scan and field-of-regard tables
    return myid
    
def _intake_original_file(db,local_pathname,name=None,keywords=None,dry_run=False,cloud_extract=None,fovs=True):
    """write the database records for a given file, excluding replica table
    db: CacheDatabase object
    local_pathname: the path at which the file can currently be accessed
    cloud_extract: function returning a dictionary (line,field,detector)->dictionary of maia cloud data, or None
    name: the unique name it will be referred to as; for IASI the filename with no suffix
    keywords: dictionary of key-values for the file
    returns database id
    """
    if name is None:
        name = os.path.split(local_pathname)[-1]
    # collect cloud mask lookup table for the granule
    if cloud_extract:
        LOG.info('invoking remote MAIA job...')
        def mask_archive(fp,name=name):
            _upload_mask_to_storage(name,fp.read())
        try:
            maia_info = cloud_extract(local_pathname, mask_func=mask_archive)
        except ValueError: # XML-RPC failure
            LOG.warning('MAIA failed! no cloud mask data available for %s' % name)
            maia_info = { }
        LOG.debug('%d records in MAIA table' % len(maia_info))
    else:
        maia_info = { }
    if cloud_extract and not fovs:
        LOG.warning("MAIA data comes in as fields of view! fovs option should be true")

    prod = iasi.open_product(local_pathname)
    # get the metadata dictionary for the whole file
    simple = iasi.simple_metadata(prod)
    simple['name'] = name  # treat native filenames as unique keys
    LOG.debug( 'simple metadata: %s' % simple)
    if not dry_run:
        myid = db.insert_simple_entry(**simple)
    else: # dummy id
        myid = uuid.uuid1()
    LOG.debug("id is %s" % myid)
    
    # insert keywords, for now without value associated (FUTURE)
    kw = dict( (x,None) for x in simple.get('keywords',[]) )
    if keywords: kw.update(keywords)
    if not dry_run:
        db.insert_keywords(myid,kw)
    else:
        LOG.debug("keywords to be inserted: %s" % kw)
        
    # get the scanline metadata generator and set it to output individual fields of view, not fields of regard
    scanner = iasi.generate_scanline_metadata(prod,fovs=fovs)
    db.insert_scanlines_from_generator(myid, scanner, maia_info, dry_run=dry_run) # fills the scan and field-of-regard tables
    return myid
    
def _intake_file(db,local_pathname,name=None,keywords=None,dry_run=False,cloud_extract=None,fovs=True):   
    "see _intake_original_file; this is a redirection layer based on filename metadata" 
    if re.match('.*?meta_IASI.*?\.p\.gz',local_pathname):
        LOG.info('found a distilled metadata file needing a cloud mask run')
        return _intake_distilled_file(db,local_pathname,name,keywords,dry_run,cloud_extract,fovs)
    else:
        return _intake_original_file(db,local_pathname,name,keywords,dry_run,cloud_extract,fovs)
        

def update_cache(db=None, use_shoplist=True, only_shoplist=False, shoplist_day=None, upload=True, dry_run=False, cloud_extract=None, max_files=None):
    """ get list of available files, shopping list, and database rows of already ingested data
    if a file is in a shopping list, send it to the cache, otherwise just note its URLs
    if cloud_extract(filename)->{ (scanline,field,detector): {cloud-data}, ... } function is included, use it
    update the database
    modified 4/2008 to include PEATE url by default
    """
    if not db: 
        db = CacheDatabase(DATABASE_URL)
    os.makedirs(TMP)
    if dry_run: upload = False
    available_files = _get_available_files()
    LOG.info( '%d available files' % len(available_files) )
    if use_shoplist:
        shoplist = shoplist_to_keywords( read_shopping_lists(shoplist_day) )
        LOG.debug( 'shopping list: %s' % shoplist )
        interesting_files = set(shoplist.keys())
    else:
        shoplist = { }
        interesting_files = set()
    if only_shoplist==True: 
        desired_files = available_files & interesting_files
    else:
        desired_files = available_files
    LOG.info( "%d desired files" % len(desired_files) )
    LOG.info( "%d cache-able files" % len(interesting_files) )
    # LOG.debug("desired files: %s" % desired_files)
    inventory = db.get_inventory()    
    for fnum, filename in enumerate(desired_files):
        if filename not in inventory: # FIXME: should also check if we have a replica in the cache, since simple may include predictive        
            LOG.info('caching #%d/%d: %s' % (fnum,len(desired_files),filename))
            # FIXME: one FTP session instead of one per file
            try:
                local_pathname = _download_file( UPSTREAM_FTP, filename )
            except:
                LOG.warning('could not download %s! skipping..' % filename)
                continue
            LOG.debug('downloaded %s as %s' % (filename,local_pathname))
            if 0 == os.stat(local_pathname).st_size: 
                LOG.warning("%s is empty! skipping" % local_pathname)
                continue
            urls = ['ftp://%s/%s/%s' % (UPSTREAM_FTP + (filename,)), peate_urimake(filename)]
            kw = dict( (k,None) for k in shoplist.get(filename,[]) ) 
            try:
                myid = _intake_file(db, local_pathname, filename, kw, dry_run, cloud_extract)
            except:
                LOG.warning("could not intake %s! putting bad-data entry in database and going on to next file" % filename)
                simple = {  'instrument': 'iasi',
                            'name': filename,
                            'keywords': ['bad', 'retry'],
                            'state': 'retry',
                            'description': 'raw IASI Level 1C data placeholder, unable to read file'
                             }
                if not dry_run and not db.contains_name(filename):
                    myid = db.insert_simple_entry(**simple)
                    _upload_to_cache(local_pathname)
                os.unlink(local_pathname)
                continue

            if not dry_run and upload and (filename in interesting_files):
                LOG.debug("uploading %s to cache" % local_pathname)
                urls += _upload_to_cache(local_pathname)
            LOG.info("replicas are at %s" % urls)
            if urls and not dry_run: 
                db.insert_replica_entries( myid, urls )
            LOG.debug('removing %s' % local_pathname)
            os.unlink(local_pathname)
            if max_files is not None: 
                max_files -= 1
                if max_files<=0: 
                    LOG.info('stopping due to max_files')
                    break
            
    os.rmdir(TMP)

def distill(db,pathnames,dry_run=False):
    """ given a list of local pathnames, generate preprocessed metadata for follow-up
    """
    # make a local directory to build stuff in
    from shutil import rmtree
    absnames = [os.path.abspath(path) for path in pathnames]
    LOG.debug(str(absnames))
    oldcwd = os.getcwd()
    TMP = os.tempnam()  # FUTURE: get rid of the warning
    os.makedirs(TMP)
    os.chdir(TMP)
    LOG.debug('working in %s' % TMP)
    # survey the files and collect the generated filenames
    zult = [_distill_file(path) for path in absnames]
    LOG.info(str(zult))
    # collect the base directory names to sync 
    candy = set( x.split(os.sep)[0] for _,x in zult )
    cmd = ['rsync', '-au', '--progress'] + list(candy) + [ SURVEY_RSYNC ]
    LOG.debug("sync command: " + str(cmd))
    if not dry_run:
        sh( cmd )
    for name,pathname in zult:
        uri = SURVEY_RSYNC+pathname
        LOG.info('follow-up: %s -> %s' % (name,uri))
        if not dry_run:
            db.update_survey(name, uri)
    os.chdir(oldcwd)
    if not dry_run:
        rmtree(TMP)
    return zult
        

def download(db,namelist,tgt_dir=None,dry_run=False):
    """
    given a list of granule names and a target directory, download the data files
    return dictionary of name=> local filename
    FUTURE: use obtain_granule to implement this
    """
    if not db: 
        db = CacheDatabase(DATABASE_URL)
    if not tgt_dir:
        tgt_dir = '.'
    uus = [ (name,db.uuid_for_name(name)) for name in namelist ]
    LOG.debug('%s' % uus)
    zult = { }
    for (name,uu) in uus:
    #for name in namelist:
        LOG.debug('downloading "%s"' % name)
        #uu = db.uuid_for_name(name)
        local = os.path.join(tgt_dir,name)
        replicas = db.urls_for_uuid(uu)
        replicas = sort_urls_by_score(replicas)
        if dry_run:
            LOG.info('replicas for %s: %s' % (name,str(replicas)))
            continue
        url = get_any_replica(replicas,local)
        if not url:
            LOG.warning('no available replicas for %s!' % name)
        else:
            LOG.info('downloaded %s from %s' % (local,url))    
            zult[name] = local
    return zult

def ingest_directory(db,pathnames,urlbase,keywords=None,upload=False,dry_run=False,cloud_extract=None,force=False):
    """
    given a list of paths corresponding to a (probably ftp://) url, e.g. using FUSE
    go through the list (filenames for IASI are unique)
    if entries already exists, skip
    get metadata for that file, and create a simple_metadata entry and a resource instance (replica) linking to the url
    """
    if dry_run: 
        upload = False
    if not db: 
        db = CacheDatabase(DATABASE_URL)
    inventory = db.get_inventory()    
    for path in pathnames:
        dirbase, filename = os.path.split(path)
        (name,) = RE_IASI.findall(filename)   # to handle meta_IASI_*.p.gz distilled files
        if name in inventory:
            if force:
                uu = db.uuid_for_name(name)
                LOG.info('removing previous entry %s for %s' % (uu,name))
                if not dry_run: 
                    db.delete_uuid(uu)
            else:
                LOG.debug('%s already in database' % name)
                continue
        myid = _intake_file(db, path, name, keywords if keywords else { }, dry_run, cloud_extract)
        if urlbase:
            urls = [urlbase % name]
        else:
            urls = []
        if upload:
            urls += _upload_to_cache(path)
        urls += [ peate_urimake(name) ]
        LOG.info('replicating %s as %s' % (name,urls))
        if urls and not dry_run: 
            db.insert_replica_entries( myid, urls )

def ingest_peate(db,year,day,keywords=None,upload=True,dry_run=False,keep=False, cloud_extract=None):
    """ screen-scrape urls from the peate database web pages
    download files locally
    add to database
    send to cache
    """
    if dry_run: 
        upload = False
    if not db: 
        db = CacheDatabase(DATABASE_URL)
    inventory = db.get_inventory()    
    urls = peate_urlget(year,day)
    LOG.debug("urls: " + str(urls))
    os.makedirs(TMP)
    for url in urls:
        (name,) = RE_IASI.findall(url)
        path = os.path.join(TMP,name)
        if name in inventory: 
            LOG.debug('skipping %s - already entered' % name)
            continue
        LOG.debug("downloading %s as %s" % (url,path))
        _get(url,path)
        myid = _intake_file(db, path, name, keywords if keywords else { }, dry_run, cloud_extract)
        urls = []
        if upload:
            urls += _upload_to_cache(path)
        urls.append( PYTHON_URI_BASE + "peate_urlget(%04d,%02d,'%s')" % (year,day,name) )
        LOG.info('replicating %s as %s' % (name,urls))
        if urls and not dry_run: 
            db.insert_replica_entries( myid, urls )
        if not keep:
            os.unlink(path)
    if not keep:
        os.rmdir(TMP)

def backfill_peate_replicas(db, start_date, end_date, dry_run=False):
    """Add PEATE replica URIs for granules during a range of days
    This is needed to catch-up to data which is now available in the PEATE
    """
    # grab an ORM session so we can access granules as objects with attributes
    S=db.session()
    # search by attribute
    granules = S.query(Granule).filter((Granule.start_time>=start_date) & (Granule.end_time<end_date)).all()
    LOG.info('checking %d granules' % len(granules))
    for gran in granules:
        print('checking %d urls for granule %s' % (len(gran.replicas),gran.name))
        found = False
        # check whether we already have a peate url in the replica list
        for url in [x.content_url for x in gran.replicas]:
            if PYTHON_URI_BASE not in url: continue
            found=True
            break
        if not found:
            uri = peate_urimake(gran.name)
            print(' + %s' % uri)
            if not dry_run: 
                db.insert_replica_entries(gran.uuid, [uri])

def backfill_uncached(db=None, upload=True, dry_run=False):
    """ find anything that's still on woodstock but doesn't have a replica entry
    transfer it over to the SAN
    """
    if not db: 
        db = CacheDatabase(DATABASE_URL)
    os.makedirs(TMP)
    if dry_run: upload = False
    available_files = _get_available_files()
    LOG.info( '%d available files' % len(available_files) )
    huntdown = db.granules_without_replicas() # dictionary of uuid -> name
    toxfer = set(huntdown.values()) & available_files
    name2uuid = dict( (y,x) for x,y in huntdown.items() )
    LOG.info('%d files to transfer and mark as replicas' % len(toxfer))  
    for filename in toxfer:
        LOG.info('caching %s' % filename)
        # FIXME: one FTP session instead of one per file
        try:
            local_pathname = _download_file( UPSTREAM_FTP, filename )
        except:
            LOG.warning('could not download %s! skipping..' % filename)
            continue
        LOG.debug('downloaded %s as %s' % (filename,local_pathname))
        urls = ['ftp://%s/%s/%s' % (UPSTREAM_FTP + (filename,))]
        if upload:
            LOG.debug("uploading %s to cache" % local_pathname)
            urls += _upload_to_cache(local_pathname)
        LOG.info("replicas are at %s" % urls)
        if urls and not dry_run: 
            db.insert_replica_entries( name2uuid[filename], urls )
        os.unlink(local_pathname)
    os.rmdir(TMP)

def backfill_distilled(db, cloud_extract=None, names=None, force=False, count=1, dry_run = False):
    """ grab an entry from survey table without an ingest date (or named entry)
    download distilled file (.p.gz) from indicated URL
    run MAIA if available
    ingest distilled metadata and MAIA cloud mask data
    """
    if names:
        LOG.debug('using list of names for backfill')
        count = len(names)
    
    for dex in range(count):
        if names: 
            entry = db.survey_entry_for_name(names[dex])
        else:
            entry = db.pick_unfinished_survey_entry()
            
        # check if the file isn't in the database, if we're not allowed to force
        if entry is None: 
            LOG.warning('no more entries available!')
            break
        LOG.info("processing %s" % entry.name)
        if db.contains_name(entry.name):
            if force:
                uu = db.uuid_for_name(entry.name)
                LOG.info('removing previous entry %s for %s' % (uu,entry.name))
                if not dry_run: 
                    db.delete_uuid(uu)
            else:
                LOG.debug('%s already in database' % entry.name)
                continue
        try:            
            # download distilled metadata file
            filename = entry.content_uri.split('/')[-1]
            _get(entry.content_uri, filename, may_decompress=False)
            
            # send the file through intake process
            _intake_file(db, filename, cloud_extract = cloud_extract, dry_run=dry_run)
            
            # delete the survey file
            os.unlink(filename)
            
        except: # we had a problem - put the entry back in the uningested list if it hadn't already been done
            if entry.ingest_time is None:
                LOG.debug('moving entry %s back to un-ingested list' % entry.name)
                db.abandon_unfinished_survey_entry(entry)
            raise
  
def check():
    """ check the files in the cache and remove any instances rows for files that have been purged
    update last_seen field in the database for each instance
    """
    raise NotImplementedError
    
def fix_iasi_field_uuids(session, granule_query=None):
    """ go through all granules and update IasiFields with uuids
    S = db.session()
    for gran,field in iasi_granule_field_pairs(S):
        field.uuid = gran.uuid
    S.commit()
    """
    if granule_query is None:
        granule_query = session.query(Granule).filter_by(instrument='iasi')
    N = granule_query.count()
    nchanged = 0
    for gran_num in range(N):
        granule = granule_query[gran_num]
        if granule.instrument != 'iasi': continue
        if (granule.start_time==None) or (granule.end_time==None): continue
        print "%05d/%05d %s" % (gran_num, N, granule)
        modded = []
        for field in session.query(IasiField).filter( 
                                (granule.start_time <= IasiField.for_time) &
                                (IasiField.for_time <= granule.end_time) ):
            if field.uuid==None:
                field.uuid = granule.uuid
                modded.append(field)
                nchanged += 1
        session.commit()
        session.flush()
    print "total number of fields changed: %d" % nchanged

def iasi_granules_for_fields(session):
    """ go through all granules and generate Granule, IasiField pairs for editing
    S = db.session()
    for gran,field in iasi_granule_field_pairs(S):
        field.uuid = gran.uuid
    S.commit()
    """
    for field in session.query(IasiField):
        try:
            (g,) = session.query(Granule).filter( 
                                (Granule.instrument=='iasi') &
                                (Granule.start_time <= field.for_time) &
                                (field.for_time <= Granule.end_time) ).all()
            yield g, field
        except ValueError:   # tuple unpack ouchie, too many or too few values
            print "no unique granule for field %s" % field


