#! /usr/bin/env python
""" $Id: iasi2fbf.py,v 1.12 2009/03/19 23:13:17 rayg Exp $
Extract IASI data to flat binary files in a similar form to AIRS ROK MATLAB script.

spectralFreq.real4.WWWW # wavenumbers
radiances.real4.WWWW # radiance data
time.real4       # time of day in decimal hours?
latitude.real4   # latitude in northward degrees
longitude.real4  # longitude in eastward degrees
scanAng.real4
landFrac.real4
elevation.real4  # -optional elevation (units?)
solar_zenith_angle # degrees
satellite_zenith_angle # degrees
solar_azimuth_angle # degrees
satellite_azimuth_angle # degrees

"""

CVSID="$Id: iasi2fbf.py,v 1.12 2009/03/19 23:13:17 rayg Exp $"

import sys
import os
import logging
import glob
import numpy
import eugene
import metahoard.iasi.tools as iasi

# logger instance for operations logs
log = logging.getLogger("iasi2fbf")

# conversion functions to get numpy arrays from iasi record fields
def real4sfromarray(x): return numpy.array(x,numpy.float32)
def int4sfromarray(x): return numpy.array(x,numpy.int32)
def real8sfromarray(x): return numpy.array(x,numpy.float64)
def real4fromvalue(x): return numpy.array([x],numpy.float32)
def real4sfromdatetimes(L): 
    return numpy.array([x.hour + x.minute/60. + x.second/3600. + x.microsecond/3.6e9 for x in L],numpy.float32)

# FUTURE: make this into a generic fbf_writer that takes the table as a constructor parameter; add to fbf toolbox
class fbf_writer(object):
    """A class capable of writing a series of records to a flat binary directory
    """
    
    _base = None # directory name which we're writing
    _ignore = [] # list/set of names which we don't want written
    _files = { } # currently open file objects
    _info = { }  # info on which fields go to which files

    def __init__(self,info,output_directory_name,comment='',ignore=[]):
        if not os.path.isdir(output_directory_name): os.makedirs(output_directory_name)
        self._info = info
        self._base = output_directory_name
        self._ignore = ignore 
        self._files = { }
        self._create_metadata(comment)
        
    def _create_metadata(self,comment):
        """ write version / tracking / log metadata to the output directory
        """
        if comment: 
            fp = file(os.path.join(self._base,'README.txt'),'wt')
            print >>fp, CVSID
            print >>fp, iasi.CVSID
            print >>fp, comment

    @staticmethod
    def endian(ext):
        "change the case of the extension string based on system endianness"
        return ext.lower() if sys.byteorder=='little' else ext.upper()

    def _create(self, stem, ext, data):
        "create a binary output file if it doesnt exist -- use array shape in filename if needed"
        if hasattr(data,'shape') and getattr(data,'size',1)>1:
            dimsum = ['.%d' % x for x in data.shape]
            dimsum.reverse()
            shape = ''.join(dimsum)
        else:
            shape = ''
        filename = os.path.join(self._base, stem + self.endian(ext) + shape)
        log.debug("creating %s..." % filename)
        return file(filename, 'wb')

    def _append(self,key,data):
        "append a record to the named output key, creating files as needed"
        stem, ext, cvt = self._info[key]
        if not stem: stem = key
        # see if we need to open the output file
        if key not in self._files: 
            fp = self._create(stem, ext, data)
            self._files[key] = fp
        else:
            fp = self._files[key]
        # reformat the data as the intended type and write the next record
        reform = cvt(data)
        log.debug('writing a record of %s' % key)
        reform.tofile(fp)

    def write_ancillary(self,stem,suffix,data):
        if stem not in self._files:
            fp = self._create(stem,suffix,data) 
            self._files[stem] = fp
            data.tofile( fp )
        
    def __call__(self,record):
        for field_name in self._info.keys():
            if field_name not in self._ignore: 
                self._append(field_name, getattr(record,field_name))


# example dictionary in cloud_dict
# (22,29,2): 
# {'clear_percent': 0.0,
#  'cloudy_percent': 100.0,
#  'day_time': 0,
#  'detector_index': 2,
#  'field_index': 29,
#  'field_of_regard_time': 14216774,
#  'layer_count': 6.0,
#  'satellite_latitude': -69.090000000000003,
#  'satellite_longitude': -38.310000000000002,
#  'satellite_zenith_angle': 56.899999999999999,
#  'scan_line': 22,
#  'surface_pressure': 978.75999999999999,
#  'surface_temp': 258.38,
#  'surface_type': 0,
#  'twometer_temp': 258.23000000000002,
#  'when': datetime.datetime(2008, 6, 5, 3, 56, 56, 774000)},

class iasi_record_fbf_writer(fbf_writer):

    _detector_number = None # which detector we're writing out, if any (None implies all-detectors)
    
    # filename and data type table, noting that keys are common between the iasi_record and this table
    # we need the data types in order to use numpy to force the raw data into the right format
    IASI_FBF_FILENAMES = {  'radiances':    (None, '.real4', real4sfromarray), # by default None implies key is used as stem
                            'latitude':     (None, '.real4', real4fromvalue),
                            'longitude':    (None, '.real4', real4fromvalue),
                            'detector_number':    (None, '.real4', real4fromvalue),
                            'scan_number':    (None, '.real4', real4fromvalue),
                            'time':         (None, '.real4', real4sfromdatetimes),
                            'epoch':        (None, '.real8', real8sfromarray),
                            'metop_zenith_angle': (None, '.real4', real4sfromarray),
                            'metop_azimuth_angle': (None, '.real4', real4sfromarray),
                            'sun_zenith_angle': (None, '.real4', real4sfromarray),
                            'sun_azimuth_angle': (None, '.real4', real4sfromarray),
                            }
    # data values only available when we merge in MAIA output dictionaries from secondary files
    CLOUD_FBF_FILENAMES = { 'clear_percent': (None, '.real4', real4sfromarray),
                            'cloudy_percent': (None, '.real4', real4sfromarray),
                            'day_time': (None, '.real4', real4sfromarray),
                            'surface_pressure': (None, '.real4', real4sfromarray),
                            'surface_temp': (None, '.real4', real4sfromarray),
                            'twometer_temp': (None, '.real4', real4sfromarray),
                            'surface_type': (None, '.real4', real4sfromarray)
                            }
    IASI_FBF_FILENAMES.update( dict( (s,(None,'.real4',real4sfromarray)) for s in iasi.CLOUD_MASK_PRODUCTS ) )

    def __init__(self,output_directory_name,detector_number,comment='',ignore=[],use_cloud=False):
        inventory = dict(self.IASI_FBF_FILENAMES)
        if use_cloud: inventory.update(self.CLOUD_FBF_FILENAMES)
        fbf_writer.__init__(self,inventory,output_directory_name,comment,ignore)
        self._detector_number = detector_number

    def write_wavenumbers(self,wnums):
        self.write_ancillary('wavenumbers','.real8',wnums)

    


def cloud_mask_dict(cloud_list):
    "convert MAIA cloud mask list of dictionaries to a dictionary of dictionaries keyed from (scanline,fieldofregard,detector) tuples"
    keys = [(x['scan_line'], x['field_index'], x['detector_index']) for x in cloud_list]
    return dict( zip(keys,cloud_list) )

def cloud_mask_load(pathname):
    "return dictionary of cloud mask data for use in iasi_tools"
    from cPickle import load
    leest = load( file( pathname, 'rb' ) )
    return cloud_mask_dict(leest)


def extract_sounding(output_directory, detector_number=None, as_scan_lines=True, use_cloud=False, lines=None, ignore=[], *filenames):
    """ iterate through all the records for a given detector for a series of files, writing them to flat binary    
    """
    log.info("creating %s..." % output_directory)
    rec_num = 0
    if detector_number is None: 
        detector_info = "all detectors"
    else:
        detector_info = "detector %d" % detector_number
    comment = """Data extraction from %s for %s""" % (`filenames`, detector_info)
    write = iasi_record_fbf_writer(output_directory, detector_number, comment, ignore=ignore, use_cloud=use_cloud)
    for filename in filenames: 
        
        filename = filename if not '*' in filename else glob.glob(filename)[0]
        log.info("processing %s..." % filename)
        
        cloud_dict = None
        if use_cloud:
            dn,fn = os.path.split(filename)
            cn = os.path.join(dn,'mask_%s.pickle' % fn)
            log.debug("reading cloud data from %s" % cn)
            assert( os.path.exists(cn) )
            cloud_dict = cloud_mask_load(cn)
                
        prod = eugene.Product(filename)
        if not as_scan_lines:
            datagen = iasi.sounder_records(prod, detector_number, lines, cloud_dict = cloud_dict)
        else: 
            datagen = iasi.sounder_scanlines(prod, detector_number, lines, cloud_dict = cloud_dict)
        for record in datagen:
            log.debug(str(record))
            write( record )
            write.write_wavenumbers(record.wavenumbers) # only does it once
            rec_num += 1
            print 'wrote %5d records..   \r' % rec_num, 
            sys.stdout.flush()
    print "\ndone!"

    
    
    
    
    
class iasi_retrieval_fbf_writer(fbf_writer):
    # FUTURE: move this information to iasi_tools - especially units and scaling factors!
    # EUGENE KEY: ( fbf-stem, fbf-type, conversion-function, scaling-factor, units, is-single-level )
    INFO = {    'ATMOSPHERIC_TEMPERATURE': ('T', '.real4', real4sfromarray, 2, 'K', False), 
                'ATMOSPHERIC_WATER_VAPOUR': ('wv', '.real4', real4sfromarray, 6, 'kg/kg', False),
                'ATMOSPHERIC_OZONE': ('oz', '.real4', real4sfromarray, 7, 'kg m-2', False),
                'INTEGRATED_OZONE': ('integrated_oz', '.real4', real4sfromarray, 7, "kg m-2", True),
                'SURFACE_TEMPERATURE': ('surface_temp', '.real4', real4sfromarray, 2, "K", False),
                'INEGRATED_N2O': ('integrated_N2O', '.real4', real4sfromarray, 7, "kg m-2", True),
                'INTEGRATED_CO': ('integrated_CO', '.real4', real4sfromarray, 7, "kg m-2", True),
                'INTEGRATED_CH4': ('integrated_CH4', '.real4', real4sfromarray, 5, "kg m-2", True),
                'INTEGRATED_CO2': ('integrated_CO2', '.real4', real4sfromarray, 3, "kg m-2", True),
                'SURFACE_EMISSIVITY': ('surface_emiss', '.real4', real4sfromarray, 4, "ratio", False),
                'FRACTIONAL_CLOUD_COVER': ('cloud_cover_fraction', '.real4', real4sfromarray, 2, "%", False),
                'CLOUD_TOP_TEMPERATURE': ('cloud_top_temp', '.real4', real4sfromarray, 2, "K", False), 
                'CLOUD_TOP_PRESSURE': ('cloud_top_press', '.real4', real4sfromarray, 0, "Pa", False), 
                'CLOUD_PHASE': ('cloud_phase', '.real4', real4sfromarray, 0, "0:none 1:liquid 2:ice 3:mixed", False),
                'SURFACE_PRESSURE': ('surface_press', '.real4', real4sfromarray, 0, "Pa", True),
                'FLG_IASICLD': ('is_cloud', '.int4', int4sfromarray, 0, "boolean", True),
                'FLG_IASICLR': ('scene_type', '.int4', int4sfromarray, 0, "clear / partly cloudy / cloudy", True),
                'FLG_ITCONV': ('is_retrieval_converged', '.int4', int4sfromarray, 0, "boolean", True),
                'FLG_IASIBAD': ('is_bad_spectra', '.int4', int4sfromarray, 0, "boolean", True),
                }

    def __init__(self, output_directory_name, comment='', ignore=[]):
        NFO = dict( (x,y[0:3]) for x,y in self.INFO.items() )
        NFO['lat'] = ('latitude', '.real4', real4sfromarray)
        NFO['lon'] = ('longitude', '.real4', real4sfromarray)
        NFO['solzen'] = ('solar_zenith_angle', '.real4', real4sfromarray)
        NFO['solaz'] = ('solar_azimuth_angle', '.real4', real4sfromarray)
        NFO['satzen'] = ('satellite_zenith_angle', '.real4', real4sfromarray)
        NFO['sataz'] = ('satellite_azimuth_angle', '.real4', real4sfromarray)
        fbf_writer.__init__(self,NFO,output_directory_name,comment,ignore)
        
    def write_levels(self,ancil):
        self.write_ancillary('T_pressure_levels','.real4',real4sfromarray(ancil['PRESSURE_LEVELS_TEMP']))
        self.write_ancillary('wv_pressure_levels','.real4',real4sfromarray(ancil['PRESSURE_LEVELS_HUMIDITY']))
        self.write_ancillary('oz_pressure_levels','.real4',real4sfromarray(ancil['PRESSURE_LEVELS_OZONE']))
        self.write_ancillary('surface_emiss_wavelengths','.real8',real8sfromarray(ancil['SURFACE_EMISSIVITY_WAVELENGTHS']))
        

class retrieval_record(object):
    pass                
                
def extract_retrieval(output_directory, ignore=[], *filenames):
    comment = """Data extraction from %s\nPressure levels are in Pa\n""" % (`filenames`)
    for (key,(stem,_,_,sf,units,_)) in iasi_retrieval_fbf_writer.INFO.items():
        comment += '%s (%s): from %s\n' % (stem,units,key)
    write = iasi_retrieval_fbf_writer(output_directory, comment, ignore=ignore)
    for filename in filenames:
        log.info('processing %s...' % filename)
        prod = iasi.open_product(filename)
        ancil = iasi.retrieval_read_ancillary(prod)
        write.write_levels(ancil) # only writes for first file
        data = retrieval_record()
        for (key,(stem,_,_,sf,units,onelevel)) in iasi_retrieval_fbf_writer.INFO.items():
            log.info("reading %s as %s..." % (key,stem))
            # read that field in for the whole file, noting that integrated quantities are single-level
            # note that INEGRATED_N2O is for real (dammit guys)
            # apply scaling factor
            field =  iasi.retrieval_read(prod, key, single_level=onelevel) 
            setattr(data, key, field.squeeze())
        data.lat, data.lon = iasi.retrieval_read_location(prod)
        data.solzen, data.satzen, data.solaz, data.sataz = iasi.retrieval_read_orientation(prod)
        log.debug("writing record...")
        write(data)



def main():
    import optparse
    usage = """usage: %prog [options] list-of-input-files
$Id: iasi2fbf.py,v 1.12 2009/03/19 23:13:17 rayg Exp $
This program extracts IASI data to flat binary files using the EUGENE toolkit.
It can output in two main geometries: record-by-record or scan-lines (-L option)
In scan-line mode, data is presented geographically contiguously.
If all detectors are requested, the detectors are geographically deinterleaved
to create 'pseudo-scan-lines', each containing two of the four detectors.
UNIX epoch time is included in the output, as well as detector number and scan number.
Unless otherwise specified, output is zero-based instead of one-based.
"""

    parser = optparse.OptionParser(usage)
    parser.add_option('-t', '--test', dest="self_test",
                    action="store_true", default=False, help="run self-tests")
    parser.add_option('-d', '--detector', dest="detector", type="int", default=None,
                    help="optional: detector number, 0..3; default is all detectors")
    parser.add_option('-L', '--lines', dest='lines', default=False, action='store_true',
                    help="optional: enable extraction of a scan line at a time (three dimensional output)")
    parser.add_option('-R', '--no-radiances', dest='no_radiances', default=False, action='store_true',
                    help="optional: disable output of radiance data")                    
    parser.add_option('-S', '--scans', dest='scans', default=None,
                    help="optional: range of scan lines to export, starting from 0 e.g. -S 0-13")
    parser.add_option('-T', '--retrieval', dest='retrieval', default=False, action='store_true',
                    help='extract from a SND format retrieval file')
    parser.add_option("-C", "--cloud", dest="read_cloud_mask_files", default=False, action='store_true',
                    help="look for mask_*.pickle files with cloud mask data (see iasi_cloud_extract)")
    parser.add_option("-o", "--output", dest="output",
                    help="required: output directory name (will be created)")
    parser.add_option('-v', '--verbose', dest="verbose",
                    action="store_true", default=False, help="enable debug output")   
    parser.add_option('-q', '--quiet', dest="quiet",
                    action="store_true", default=False, help="only warning and error output")   
    (options, args) = parser.parse_args()

    lvl = logging.INFO
    if options.verbose: lvl = logging.DEBUG
    elif options.quiet: lvl = logging.WARNING
    log.setLevel(lvl)

    if options.self_test:
        import doctest
        doctest.testmod()
        sys.exit(2)

    if len(args) < 1 or not options.output:
        parser.error("incorrect arguments. try --help or -h")
    else:
        if options.scans is not None: 
            lal = [int(x) for x in options.scans.split('-')]
            first,last = lal[0], lal[-1]
            scans = range(first,last+1)
        else:
            scans = None        
        ignore = [ 'radiances' ] if options.no_radiances else []
        if not options.retrieval:
            extract_sounding( options.output, options.detector, options.lines, options.read_cloud_mask_files, scans, ignore, *args )
        else:
            assert( options.read_cloud_mask_files != True )
            extract_retrieval( options.output, ignore, *args )

if __name__=='__main__':
    logging.basicConfig()
    main()
