#!/usr/bin/env python
# encoding: utf-8
"""
iasi_cache.py

Created by Keoni on 2007-11-29.
Copyright (c) 2007 University of Wisconsin SSEC. All rights reserved.

Setting up a cache:
Create database
initialize database
create IASI_cache directory on SAN
get upstream directory set up

ssh -L5432:bora2:5432 ash
python -m iasi_cache --iasi-db=postgres://rayg:c@nnab1s@localhost/iasi_cache 

requires miscellaneous routines from keoni 0.3.4 or better (see larch.ssec.wisc.edu)

"""

_CVS_IDENT = "$Id: iasi_cache.py,v 1.124 2008/08/18 21:38:21 rayg Exp $"

import sys
import os
import logging
import uuid
import re
import time
from cPickle import dump
from ftplib import FTP
from subprocess import check_call as sh
from datetime import datetime, timedelta
from time import gmtime, strftime
from collections import defaultdict

from sqlalchemy import *    
import sqlalchemy.orm as orm

LOG = logging.getLogger( __name__ )

from metahoard.iasi import tools as iasi
from metahoard.iasi.config import *
from metahoard.iasi.schema import *
from metahoard.iasi.transfer import *
from metahoard.iasi.commands import *
from metahoard.iasi.util import *



#################
#
#   S E L F  
#  T E S T S
#
#

        
TEST_GRANULES = ['IASI_xxx_1C_M02_20080209170855Z_20080209171159Z_N_O_20080209185150Z__20080209185625']
TEST_OUTPUT_DIR = '/tmp/iasi_cache_test_' + os.environ['USER']
TEST_CONSTRAINTS_1 = [ '20071209:194215-20071209:194300' ]
TEST_CONSTRAINTS_2 = [ '20080420:120000-20080420:130000', '32o0m30sN 44W 35d17m14sS-47.25E' ]
TEST_CONSTRAINTS_3 = [ '20071204-20071205', '163.146E 25.474N 181.147E 30.021N' ]
TEST_CONSTRAINTS_4 = [ '20080422:165000-20080422:170000', '-', 'conus', 'fields=14,15', 'cloudy_percent<5', 'clear_percent>95', 'day_time==2', 'surface_type==0']

TEST_CONSTRAINTS_ALL = [ TEST_CONSTRAINTS_1, TEST_CONSTRAINTS_2, TEST_CONSTRAINTS_3, TEST_CONSTRAINTS_4 ]

class test_file(object):
    def __init__(self,pathname):
        self._pathname = os.path.abspath(pathname)
    def __repr__(self):
        return '<test file "%s">' % self._pathname
    def __str__(self):
        return self._pathname
    def __del__(self):
        if os.path.isdir(self._pathname):
            print "removing test directory %s" % self._pathname
            os.rmdir(self._pathname)
        else:
            print "removing test file %s" % self._pathname
            os.unlink(self._pathname)
    
def test_keywords(db=None):
    db = db or CacheDatabase()
    print "KEYWORDS test"
    D = db.all_keywords()
    print list(D.keys())
    
def test_query(db=None):
    db = db or CacheDatabase()
    print "QUERY test"
    for constraints_strings in TEST_CONSTRAINTS_ALL:
        constraints = _constraints_parse(constraints_strings)
        print '/' * 32 + str(constraints_strings)
        query(db,**constraints)

def test_download(db=None):
    db = db or CacheDatabase()
    print "DOWNLOAD test"
    pathdict = download(db,TEST_GRANULES,dry_run=False)
    return [test_file(path) for name,path in pathdict.items()]

def test_distill(db=None,test_files=None):
    db = db or CacheDatabase()
    print "DISTILL test"
    if not test_files:
        test_files = test_download(db)
    zult = distill(db, [str(x) for x in test_files], dry_run=True)
    return [test_file(z) for y,z in zult] 

def test_export(db=None):
    db = db or CacheDatabase()
    print "EXPORT test"
    selection = { None: [14,15], (0,2,4,6): None, 11:[0,14,15,29] }
    export( db, TEST_OUTPUT_DIR, TEST_GRANULES[0], selection, keep=False )
    from shutil import rmtree
    print "export test output"
    sh(['ls', '-l', TEST_OUTPUT_DIR])
    rmtree(TEST_OUTPUT_DIR)

def test_emit(db=None):
    db = db or CacheDatabase()
    print "EMIT test"
    constraints = _constraints_parse(TEST_CONSTRAINTS_4)
    emit(db, TEST_OUTPUT_DIR, keep=True, **constraints)
    from shutil import rmtree
    print "emit test output"
    sh(['ls', '-l', TEST_OUTPUT_DIR])
    rmtree(TEST_OUTPUT_DIR)
    
def test_update(db=None):
    db = db or CacheDatabase()
    print "UPDATE test"
    update_cache(db,dry_run=True, 
                    only_shoplist=False, 
                    use_shoplist=True, 
                    shoplist_day=None,
                    cloud_extract=None, 
                    max_files=3)

def test_merge_maia(db=None):
    print "MERGE-MAIA test"
    db = db or CacheDatabase()
    ms = _connect_maia('monkey:7890')
    os.makedirs(TMP)
    merge_maia(db, TEST_GRANULES[0], ms, force_reload=True, keep=False, dry_run=True)
    from shutil import rmtree
    rmtree(TMP)
    
def test_all(db=None):
    db = db or CacheDatabase()
    for func in [ test_keywords, test_query, test_download, test_distill, test_export, test_emit, test_update, test_merge_maia ]:
        print '-'*72
        if func == test_download:
            test_granules = test_download(db)
        elif func == test_distill:
            test_distill(db,test_granules)
        else:
            func(db)
    

#################
#
#   M A I N 
#
#

def _connect_maia(hostport = 'monkey:7890'):
    """return a cloud_extract callable for an XMLRPC server running iasi_cloud_extract
    if hostport is 'localhost' or '-' then use MAIA on localhost
    """
    if hostport in ['-', 'localhost']:
        from iasi_cloud_extract import mask_extract
        monkey = mask_extract
    else:
        host,port=hostport.split(':')
        port = int(port)
        from iasi_cloud_extract import remote_mask_extract
        monkey = remote_mask_extract( (host,port) )
    def monkeyshine(filename, text_data=None, **kwargs):
        return dict( ((x.scan_line, x.field_index, x.detector_index), vars(x)) for x in monkey(filename,text_data,**kwargs) )
    return monkeyshine
    
def _constraints_parse(args):
    """ build dictionary of strings for common orderings of constraint arguments
    used for emit and query
    FUTURE: consider turning this into a search_constraints class of similar nature to optparse, 
    combining in with _ui_query; CacheDatabase would be a factory for these objects, and accept them as input
    """
    D = {'startend': args[0]}
    D['box'] = args[1] if len(args)>1 else None
    D['keywords'] = args[2] if len(args)>2 else None
    # field-of-view expressions to be evaluated into where-expressions
    str_con = [] 
    D['string_criteria'] = str_con   # field criteria
    # anything afterward is constraint=condition
    # FUTURE: make everything c=c
    for arg in args[3:]:
        if 'fields' in arg:
            eqof = arg.index('=')
            k,v = arg[:eqof], arg[eqof+1:]
            D[k] = v
        else:
            str_con.append(arg)
    LOG.debug("constraints: %s" % str(D))
    return D

def main():
    logging.basicConfig()
    import optparse
    usage = """usage: %prog [options] action parameters
$Id: iasi_cache.py,v 1.124 2008/08/18 21:38:21 rayg Exp $

See http://katamari.ssec.wisc.edu/resources/iasi/iasi-cache-usage-notes for more information.

data user actions:

    keywords
        list keywords present in database
        (keywords are currently stored on a per-granule basis)
 
    query YYYYMMDD:HHMMSS-YYYYMMDD:HHMMSS {BBOX} {keyword1,keyword2} {fields=#,#,#} {cloudy_percent<X} ...
        list fields-of-regard fitting a given time range and bounding box
        requires a time, bounding box, or both
        to omit time or bounding box, replace with a dash (-)
        bounding box can be of the form '###N ###E ####N ###E' - see below for full coord syntax details
        keywords are optional and are use to constrain to only granules having that keyword
        fields= constraint is also optional and lists fields-of-regard to limit to 
        e.g. fields=14,15 for center-scan
        field-of-regard constraint clauses can include operators ==, <, >, <=, >= and values
        'surface_type', 'clear_percent', 'cloudy_percent', 'surface_temp', 
        'twometer_temp', 'surface_pressure', 'day_time', 'detector'
        day_time values are 0 for night, 1 for twilight, 2 for day, 3 for sun glint
        surface_type values are 0 for sea, 1 for land, 2 for coast        
        
    download granule-name1 {granule-name2 ...}
        given the full name of one or more granule files (e.g. from query or keywords)
        download the native data file to the current directory
        
    emit /path/to/target/directory YYYYMMDD:HHMMSS-YYYYMMDD:HHMMSS {BBOX} {keyword1,keyword2} {fields=#,#,#} {cloudy_percent<X} ...
        write MAT files for fields-of-regard fitting a given time range and bounding box
        creates the target directory if necessary
        requires a time, bounding box, or both
        to omit either time or bounding box, replace with a dash (-)
        constraint expressions are identical in order and content to 'query' command - see above

    export /path/to/target/directory granule-name 'line0: [field0, field1, field2, ...], line1: [ field0, ... ]'
        download the named granule
        write MAT files for the named fields-of-regard
        use ALL if all scan lines, or all fields should be used
        multiple scan lines numbers can be enclosed in ()
        all scan line and field numbers start with 0 (per EUGENE)
        examples: 
        export /tmp/mydata IASI_xxx... '0:[1,4,9], 4:[2,3,4,9], 19:[7]'
        export /tmp/mydata IASI_xxx... 'ALL: [13,14,15]'
        export /tmp/mydata IASI_xxx... 'ALL: ALL'
        export /tmp/mydata IASI_xxx... '(1,2,4): ALL'
        export /tmp/mydata IASI_xxx... 'ALL: ALL, 11: [19,20]'

BBOX - specifiying bounding boxes
(lon,lat),(lon,lat) can be used, or a human-readable format for two box corners: 
""" + COORDINATE_FORMAT_DOC + """


cache maintainer actions:   

    available
        list available files on woodstock

    shoplist 
        download shopping list from monkey and list filenames and zones

    update 
        read data from woodstock, cache it on the SAN, store replica information
        uses shopping list from monkey if available

    ingest URL-FORMAT keyword1,keyword2 file1 file2 file3 file4 ...
        add local IASI files to the database and upload to the cache
        URL-STEM is an optional URL format string (e.g. http://peate/get-iasi.cgi?%s) 
            which specifies where the files can be downloaded
            use '-' or '' as a non-value for the URL-STEM

    merge-maia granule-name
        using a granule filename or granule name, download granule if necessary
        and update field table to include MAIA output.
        replaces scans and fields for that granule

    backfill
        find granules which have no replicas listed in the database
        if the files are available on woodstock, send them to the cache 

    backfill_peate_replicas startYYYYMMDD endYYYYMMDD
        add peate URL replicas to granules in the given date range
        if they're not already present
        
    associate granule-name url1 url2 ...
        associate access URIs with a granule
        if a url is given as @PEATE, a PEATE uri is inserted

    ingest-peate YYYY DDD keyword1,keyword2
        download files from the PEATE storage system and cache them
        keywords are optional
        
    distill filename1 {filename2 ...}
        create survey files and upload to SAN storage; add to survey database table
        survey files (meta_*.p) are distilled metadata from IASI raw files.
        these files can be picked up with ingest-survey and used to generate cloud mask
        data using MAIA, as well as to fill out search database. 
        In other words, this is IASI file preprocessing for the PEATE storage system.
        Survey files are then routed to zara for cloud mask and database table generation. 

    backfill-distilled [granule-count or granule-name1] {granule-name2 ...}
        backfill distilled data using pending "survey" table
        When "distill" command is run on a granule, it logs a survey entry
        and puts a distilled metadata file on the SAN. 
        This command recovers those distilled metadata files, 
        optionally runs MAIA to extract cloud masks (-M option), 
        and then updates the database metadata.
        If invoked with --force, previous database records will be 
        overwritten.
        
            
"""

    parser = optparse.OptionParser(usage)
    parser.add_option('-t', '--test', dest="self_test",
                    action="store_true", default=False, help="run self-tests")
    parser.add_option("-d", "--dry-run", dest="dry_run", action='store_true',
                      help="do not move/copy files or do database inserts")    
    parser.add_option("-k", "--keep", dest="keep", action='store_true',
                      help="do not delete raw data file after downloading for export or emit")
    parser.add_option("-S", "--no-shoplist", dest="no_shoplist", action='store_true',
                      help="don't use shopping lists from monkey.ssec.wisc.edu for update")    
    parser.add_option("-s", "--only-shoplist", dest="only_shoplist", action='store_true',
                      help="only index granules listed in shopping lists. default is to index all, cache on SAN only if on shopping list")
    parser.add_option("-f", "--force", dest="force", action='store_true',
                      help="backfill-distilled: where previous data exists, overwrite it")
    parser.add_option("-v", "--verbose", dest="debug", action='store_true',
                      help="turn on optional debug output")    
    parser.add_option("-q", "--quiet", dest="quiet", action='store_true',
                      help="only print errors and warnings; assume 'no' on mailing crash dumps to rayg")
    parser.add_option("-D", "--shoplist-day", dest="yyyymmdd", 
                      help="YYYYMMDD of shopping list to use, if any")    
    parser.add_option("-I", "--iasi-db", dest="db_url", 
                      help="optional URL of database to use")            
    parser.add_option("-M", "--maia-server", dest="maia_server", 
                      help="host:port of XMLRPC server for MAIA conversion of cluster data to cloud mask")            
    # parser.add_option("-f", "--force", dest="force", action='store_true',
    #                   help="force merge command to reload scanlines and fields")            
                      
    (options, args) = parser.parse_args()
    if options.self_test:
        # import doctest
        # doctest.testmod()
        test_all(CacheDatabase(options.db_url))
        sys.exit(2)
    lvl = logging.INFO
    if options.quiet: lvl = logging.WARNING
    if options.debug: lvl = logging.DEBUG
    LOG.setLevel(lvl)
    if options.yyyymmdd: 
        import time
        # convert to datetime object
        options.yyyymmdd = datetime.strptime( options.yyyymmdd, '%Y%m%d' )
    db_url = options.db_url if options.db_url else DATABASE_URL
    if db_url=='localhost': db_url=TEST_DATABASE_URL

    if len(args)==0:
        parser.error("incorrect arguments. try --help or -h")
        return 5
        
    # configure XMLRPC server to provide cloud mask records
    if options.maia_server:
        cloud_extract = _connect_maia(options.maia_server)
    else:
        cloud_extract = None

    # FUTURE: use the docstring from each of the do_ routines to generate the overall help text!
    CMDS = { }    
    
    def do_available():
        from pprint import pprint
        lal = _get_available_files()
        print '\n'.join( lal )
        print "%d files" % len(lal)
    CMDS['available'] = do_available
    
    def do_update(): 
        update_cache(dry_run=options.dry_run, 
                    only_shoplist=options.only_shoplist, 
                    use_shoplist=not options.no_shoplist, 
                    shoplist_day=options.yyyymmdd,
                    cloud_extract=cloud_extract)
    CMDS['update'] = do_update

    def do_distill():
        db = CacheDatabase(db_url)
        distill(db, args[1:], dry_run=options.dry_run)
    CMDS['distill'] = do_distill

    def do_backfill_distilled():
        db = CacheDatabase(db_url)
        if len(args)==2 and re.match('\d+',args[1]):
            count = int(args[1])
            names = None
            LOG.debug('about to backfill %d granules' % count)
        else:
            count = 1
            names = args[1:]
        backfill_distilled( db, 
                            cloud_extract = cloud_extract,
                            names = names, 
                            count = count, 
                            force = options.force, 
                            dry_run = options.dry_run 
                            )
    CMDS['backfill-distilled'] = do_backfill_distilled

    def do_ingest():
        db = CacheDatabase(db_url)
        filelist = args[3:]
        if args[2] not in [ '', '-' ]:
            kw = dict( (x.strip(),None) for x in args[2].split(',') )
        else:
            kw = { }
        ingest_directory(db, filelist, args[1] if args[1]!='-' else None, keywords=kw, dry_run=options.dry_run, force=options.force, cloud_extract=cloud_extract)
    CMDS['ingest'] = do_ingest

    def do_ingestpeate():
        db = CacheDatabase(db_url)
        year,day = map( int, args[1:3] )
        kw = dict( (x.strip(),None) for x in args[3].split(',') ) if len(args)>=4 else None
        ingest_peate(db, year, day, kw, dry_run=options.dry_run, keep=options.keep, cloud_extract=cloud_extract)
    CMDS['ingest-peate'] = do_ingestpeate
    
    def do_backfill():
        db = CacheDatabase(db_url)
        backfill_uncached(db, dry_run=options.dry_run)
    CMDS['backfill'] = do_backfill
    
    # def do_backfill_field_uuids():
    #     db = CacheDatabase(db_url)
    #     db.iasi_backfill_field_uuids(dry_run=options.dry_run)
    # CMDS['field_uuids'] = do_backfill_field_uuids
    
    def do_query():
        db = CacheDatabase(db_url)
        # time_text = args[1]
        # box_text = args[2] if len(args)>2 else None
        # keyword_list = args[3] if len(args)>3 else None
        constraints = _constraints_parse(args[1:])
        query(db, **constraints ) 
    CMDS['query'] = do_query
    
    def do_download():
        db = CacheDatabase(db_url)
        download(db,args[1:],dry_run=options.dry_run)
    CMDS['download'] = do_download
    
    def do_emit(): 
        db = CacheDatabase(db_url)
        tgt_dir = args[1]
        # time_text = args[2]
        # box_text = args[3] if len(args)>3 else None
        # keyword_list = args[4] if len(args)>4 else None
        constraints = _constraints_parse(args[2:])
        emit(db, tgt_dir, keep=options.keep, **constraints)
    CMDS['emit'] = do_emit

    def do_export(): 
        db = CacheDatabase(db_url)
        tgt_dir = args[1]
        gran_name = args[2]
        loclz = dict( ALL=None, all=None, ALLFIELDS=range(30), CENTER=range(14,16), center=range(14,16) )
        toexport = eval('{' + args[3] + '}', loclz)
        export( db, tgt_dir, gran_name, toexport, keep=options.keep )
    CMDS['export'] = do_export
    
    # def merge_maia(db, granule_name, cloud_extract_func, force_reload = False, keep = False, fovs=True, dry_run=False):
    def do_merge_maia():
        db = CacheDatabase(db_url)
        for gran_name in args[1:]:
            merge_maia(db, gran_name, cloud_extract, force_reload=True, keep=options.keep, dry_run=options.dry_run)
    CMDS['merge-maia'] = do_merge_maia
    
    def do_backfill_peate_replicas():
        db = CacheDatabase(db_url)
        start = datetime_from_string(args[1])
        end = datetime_from_string(args[2])
        print "backfilling peate replicas from %s to %s" % (str(start),str(end))
        backfill_peate_replicas(db, start, end, dry_run=options.dry_run)
    CMDS['backfill_peate_replicas'] = do_backfill_peate_replicas
    
    def do_associate():
        db = CacheDatabase(db_url)
        name = args[1]
        raw_urls = args[2:]
        urls = [peate_urimake(name) if '@PEATE'==url else url for url in raw_urls] 
        uu = db.uuid_for_name(name)
        print "found uuid %s" % str(uu)
        preexisting = db.urls_for_uuid(uu)
        toadd = [x for x in urls if x not in preexisting]
        print "adding %s\nto %s" % (toadd, preexisting)
        if not options.dry_run: 
            db.insert_replica_entries(uu, toadd)
        else:
            print "dry run - not inserting"
        print 'done!'
    CMDS['associate'] = do_associate
    
    def do_keywords():
        db = CacheDatabase(db_url)
        D = db.all_keywords()
        # sort by number of granules
        V = [ (len(entries), key) for (key,entries) in D.items() ]
        V.sort(reverse=True if not options.debug else False)
        for N,K in V:
            if len(args)>1 and K not in args: continue
            print '%16s: \t%d granules' % (K,N)
            if len(args)>1:
                for uu in D[K]:
                    print db.name_for_uuid(uu)
    CMDS['keywords'] = do_keywords
    
    def do_shoplist():
        slist = shoplist_to_keywords( read_shopping_lists( options.yyyymmdd ) )
        for fn,zone in slist.items(): 
            print fn,'\t',', '.join(zone)
    CMDS['shoplist'] = do_shoplist
    
    def do_test_exception():
        "force an exception for debug dump testing"
        raise NotImplementedError
    CMDS['exception'] = do_test_exception
    
    try:
        cmd = CMDS[args[0]]
    except:
        parser.error('command not found: should be in %s' % CMDS.keys())
        return 5
    try:
        cmd()
    except:
        if (not options.quiet) and (sys.stdin.isatty()) and (raw_input("EXCEPTION: mail crash dump to rayg? [y/n] ").strip() in 'yY'):
            import traceback
            from subprocess import call
            body = 'OPTIONS: ' + str(options) +'\n'+'ARGS: '+ str(args) + '\n\nEXCEPTION LOG:\n' + traceback.format_exc() + '\n\nSTACK TRACE:\n' + ''.join(traceback.format_stack())
            stdi = os.tmpfile()
            stdi.write(body)
            stdi.seek(0)
            call( ['mail', '-s', 'auto: iasi_cache exception', 'rayg@ssec.wisc.edu'], stdin=stdi )
            print "CRASH DUMP mailed to rayg"
        raise



    
if __name__ == '__main__':
    logging.basicConfig()
    LOG.setLevel(logging.INFO)
    sys.exit(main())
