#!/usr/bin/env python
# encoding: utf-8
"""
iasi_cache.py

Created by Keoni on 2007-11-29.
Copyright (c) 2007 University of Wisconsin SSEC. All rights reserved.

Setting up a cache:
Create database
initialize database
create IASI_cache directory on SAN
get upstream directory set up

ssh -L5432:bora2:5432 ash
python -m iasi_cache --iasi-db=postgres://rayg:c@nnab1s@localhost/iasi_cache 

requires miscellaneous routines from keoni 0.3.4 or better (see larch.ssec.wisc.edu)

"""

_CVS_IDENT = "$Id: iasi_cache.py,v 1.124 2008/08/18 21:38:21 rayg Exp $"

import sys
import os
import logging
import uuid
import re
import time
from cPickle import dump
from ftplib import FTP
from subprocess import check_call as sh
from datetime import datetime, timedelta
from time import gmtime, strftime
from collections import defaultdict

from sqlalchemy import *    
import sqlalchemy.orm as orm

LOG = logging.getLogger( __name__ )

from metahoard.iasi import tools as iasi
from metahoard.iasi.config import *

from keoni.map.parse import parse_coords, format_coords, COORDINATE_FORMAT_DOC   # see larch.ssec.wisc.edu
from keoni.time.epoch import datetime_from_string
from keoni.database.pg_alchemy import *
# try:
#     from iasi_cloud_extract import remote_mask_extract # uses MAIA to get cloud information for (scanline,field,detector) -> record
# except:
#     LOG.error('installation problem: unable to load iasi_cloud_extract!')
#     def cloud_extract(local_pathname):
#         LOG.warning('iasi_cloud_extract unavailable')
#         return { }


# ORM class stubs
class Granule(object):
    def __repr__(self):
        return '<granule %s>' % repr(self.name)
    
class KeyValue(object):
    def __repr__(self):
        return "{%s: %s}" % (repr(self.key), repr(self.value))
    
class Replica(object):
    def __repr__(self):
        return repr(self.content_url)
    
class IasiScan(object):
    def __repr__(self):
        return "<iasi scan #%d @ %s~%s>" % (self.scan_index, self.start_time, self.end_time)
    
class IasiField(object):
    def __repr__(self):
        return "<iasi field of regard @ %s>" % self.for_time

class IasiSurvey(object):
    pass
    

class CacheDatabase(object):
    """abstraction of data cache tables
    resource_simple_metadata
        uuid : granule UUID, unique
        name ; String, unique, original name of the file
        instrument: string
        perimeter: polygon
        start: datetime
        end: datetime
        span: lseg  (y is always 0, x is unix epoch time, spans from start to end)
        contact: string, email address
        state: string
        description: string
    resource_instance -- several instances may represent the same uuid
        id: integer serial
        uuid : granule UUID, foreign key is the uuid in simple_metadata
        content_url : text      # url of the actual data, including access protocol
        protocol: string        # since some http urls are actually other data formats.. e.g. dap, rsync, svn, 
        incept: timestamp   # when this resource was added to the replicas
        # optional fields
        quicklook_url: string   # http url where further information on this can be found
        expiry: timestamp   # when we stop using this resource
        seen: timestamp     # last time its existence was checked
    resource_keywords:
        id: integer serial
        uuid: granule UUID foreign keyed to simple_metadata
        keyword: String
        value: String, optional
    iasi_scan_metadata:
        id: integer serial
        uuid: granule UUID
        scan_time: timestamp, truncated to seconds, first time in scan
        scan_index: integer, index within the source file (i.e. record number counting from 0)
        perimeter: polygon, bounds around all four detectors for the scan line
        start_time: datetime, start time including microseconds
        end_time: datetime, end time including microseconds
        span: lseg, line segment of (unix-epoch,0) with double precision epoch time including microseconds
        moacvr: 30x1 string of 0s and 1s for each of the 30 fields-of-regard in the scan line DEPRECATED
    iasi_field_of_regard_navigation:
        uuid: varchar(32)
        id: integer serial
        for_time: timestamp
        for_center: point, (lon, lat) footprint center
        for_index_within_scan: integer, 0..29, which field of regard within its scan
        for_detector_number: 0..3 detector number, None for fields-of-regard (older rows)
        surface_type: smallint, MAIA surface type
        clear_percent: smallint, MAIA percent clear
        cloudy_percent: smallint, MAIA percent cloudy
        surface_temp: smallint, MAIA surface temp in Kelvin
        twometer_temp: smallint, MAIA two-meter temp in Kelvin
        surface_pressure: smallint, MAIA surface pressure in millibars
        day_time: smallint, MAIA time of day
    iasi_survey: holds information on 'raw' data which gets processed to build database entries for FOVs and granules
        id: integer serial
        name: iasi raw data filename, unique
        creation_time: datetime, nonnull, creation time of the survey pickle file
        content_uri: where to find the survey pickle file
        ingest_time: time when pickle was put in main searchable database
    """
    
    TBL_SIMPLE_METADATA = "resource_simple_metadata"
    TBL_INSTANCE = "resource_instances"
    TBL_KEYWORDS = "resource_keywords"
    TBL_SCAN_METADATA = 'iasi_scan_metadata'
    TBL_FOR_NAV = "iasi_field_of_regard_navigation"
    TBL_SURVEY = "iasi_survey"
    
    _t_simple = None  # table of simple information on a resource (file)
    _t_replica = None # table of replicas for a given resource uuid (available download urls)
    _t_scan = None # table of scan metadata, each having 30 fields-of-regard containing 120 spectra
    _t_for = None # table of fields of regard times and footprint locations
    _t_survey = None # table representing metadata that hasn't been fully prepared for database yet (unprocessed clusters etc)
    
    def schema(self, metadata=None, create=False):
        """initialize a metadata schema; optionally creating tables in a postgresql database
        """
        # FUTURE: ditch keyword columns in favor of a uuid-keyword table??
        if not metadata: metadata=self._metadata
        self._t_simple = Table( self.TBL_SIMPLE_METADATA, metadata, 
            Column( 'uuid', Uuid, primary_key=True, unique=True ),
            Column( 'name', String(), unique=True, nullable=False ),  # the original name of the resource
            Column( 'instrument', String(32) ),
            Column( 'perimeter', Polygon ),
            Column( 'start_time', DateTime ),
            Column( 'end_time', DateTime ),
            Column( 'span', LSeg ),
            Column( 'contact', String() ),
            Column( 'state', String(64) ),
            Column( 'description', String )
            )
        self._t_replica = Table( self.TBL_INSTANCE, metadata, 
            Column( 'id', Integer, primary_key=True), 
            Column( 'uuid', Uuid, ForeignKey( self.TBL_SIMPLE_METADATA + '.uuid' ) ),
            Column( 'content_url', String(), unique=True, nullable=False ),  # unique?
            Column( 'protocol', String ),
            Column( 'quicklook_url', String() ),
            Column( 'incept', DateTime, nullable=False ),
            Column( 'expiry', DateTime ),
            Column( 'seen', DateTime )
            )
        self._t_keywords = Table( self.TBL_KEYWORDS, metadata, 
            Column( 'id', Integer, primary_key=True), 
            Column( 'uuid', Uuid, ForeignKey( self.TBL_SIMPLE_METADATA + '.uuid' ) ),
            Column( 'key', String(), nullable=False ),
            Column( 'value', String() )  # FUTURE: make this a pickletype or a JSON string!!!!!!!
            )
        self._t_scan = Table( self.TBL_SCAN_METADATA, metadata, 
            Column( 'id', Integer, primary_key=True), 
            Column( 'uuid', Uuid, ForeignKey( self.TBL_SIMPLE_METADATA + '.uuid' ) ),
            Column( 'scan_time', DateTime ), # truncated to the second
            Column( 'scan_index', Integer ), # index within the file
            Column( 'perimeter', Polygon ),
            Column( 'start_time', DateTime ), # includes microseconds
            Column( 'end_time', DateTime ),
            Column( 'span', LSeg ),  # epoch time,0 line segment for intersection calculations
            Column( 'moacvr', String(30) )  # 0's or 1's from moacvr files indicating whether a FOV is of interest
            )
        self._t_for = Table( self.TBL_FOR_NAV, metadata, 
            Column( 'uuid', Uuid ), # FIXME , ForeignKey( self.TBL_SIMPLE_METADATA + '.uuid' ) ),
            Column( 'id', Integer, primary_key=True, unique=True, nullable=False ),
            Column( 'for_time', DateTime, nullable=False ), 
            Column( 'for_center', Point, nullable=False ),
            Column( 'for_index_within_scan', Integer, nullable=False ),  
            Column( 'for_detector_number', SmallInteger ),
            Column( 'surface_type', SmallInteger ),   # provided by MAIA from here on
            Column( 'clear_percent', SmallInteger ),
            Column( 'cloudy_percent', SmallInteger ),
            Column( 'surface_temp', SmallInteger ),
            Column( 'twometer_temp', SmallInteger ),
            Column( 'surface_pressure', SmallInteger ),
            Column( 'day_time', SmallInteger )
            )
        self._t_survey = Table( self.TBL_SURVEY, metadata,
            Column( 'id', Integer, primary_key=True), 
            Column( 'name', String(), unique=True, nullable=False ), 
            Column( 'creation_time', DateTime, nullable=False ), 
            Column( 'content_uri', String() ),
            Column( 'ingest_time', DateTime )
            )
        if create:
            self._t_simple.create()
            self._t_replica.create()
            self._t_keywords.create()
            self._t_scan.create()
            self._t_for.create()
            self._t_survey.create()

    # these keys can be used to constraint expressions pulled in from the command line
    FIELD_OF_VIEW_KEYS = ['surface_type', 'clear_percent', 'cloudy_percent', 'surface_temp', 'twometer_temp', 'surface_pressure', 'day_time']

    def connect(self, db_url, create=False):
        self._engine = create_engine(db_url)
        self._metadata = MetaData()
        self._metadata.bind = self._engine
        self._connection = self._engine.connect() # needed for transactions
        self.schema(create=create)
        
    def create_mappings(self,lazy=True):
        """create object-relational mappings objects to allow Granule objects to function
        """
        self._keyvalue_mapper = orm.mapper(KeyValue, self._t_keywords)
        self._replica_mapper = orm.mapper(Replica, self._t_replica)
        self._field_mapper = orm.mapper(IasiField, self._t_for, properties={
                    'scan': orm.relation(IasiScan, 
                                primaryjoin=( (self._t_scan.c.start_time <= self._t_for.c.for_time) & 
                                            (self._t_for.c.for_time <= self._t_scan.c.end_time) ),
                                viewonly=True,
                                lazy=lazy,
                                foreign_keys=[self._t_scan.c.start_time, self._t_scan.c.end_time]),
                    'granule': orm.relation(Granule, 
                                primaryjoin=( (self._t_simple.c.instrument=='iasi') & 
                                            (self._t_simple.c.start_time <= self._t_for.c.for_time) & 
                                            (self._t_for.c.for_time <= self._t_simple.c.end_time) ),
                                viewonly=True,
                                lazy=lazy,
                                foreign_keys=[self._t_simple.c.start_time, self._t_simple.c.end_time]) 
                                })
                                
        self._scan_mapper = orm.mapper(IasiScan, self._t_scan, properties={
                    'fields': orm.relation(IasiField, #backref='scan',
                                foreign_keys=[self._t_for.c.for_time], # self._t_scan.c.start_time, self._t_scan.c.end_time],
                                primaryjoin=( (self._t_scan.c.start_time <= self._t_for.c.for_time) & 
                                              (self._t_for.c.for_time <= self._t_scan.c.end_time) ), 
                                viewonly=True,
                                cascade="all, delete, delete-orphan",
                                lazy=True)
                                })

        self._entry_mapper = orm.mapper(Granule, self._t_simple, properties={    
                    'keywords': orm.relation(KeyValue, backref='granule',
                                primaryjoin=self._t_keywords.c.uuid==self._t_simple.c.uuid, 
                                lazy=lazy,  # only load when attribute is accessed
                                cascade="all, delete, delete-orphan"),
                    'replicas': orm.relation(Replica, backref='granule', 
                                primaryjoin=self._t_replica.c.uuid==self._t_simple.c.uuid,                                 
                                lazy=lazy,
                                cascade="all, delete, delete-orphan"),
                    'scans': orm.relation(IasiScan, backref='granule', 
                                primaryjoin=self._t_scan.c.uuid==self._t_simple.c.uuid,
                                lazy=lazy,                           
                                cascade="all, delete, delete-orphan"),
                    # 'fields': orm.relation(IasiField, backref="granule",
                    #             # cascade="all, delete, delete-orphan",
                    #             primaryjoin=( (self._t_simple.c.start_time <= self._t_for.c.for_time) & 
                    #                         (self._t_for.c.for_time <= self._t_simple.c.end_time) & 
                    #                         (self._t_simple.c.instrument=='iasi') ),
                    #             lazy=lazy,
                    #             viewonly=True,
                    #             foreign_keys=[self._t_simple.c.start_time, self._t_simple.c.end_time]) # needed because some uuids are NULL so we can't constrain to foreignkey in the database yet
                    }
                )
                
        self._survey_mapper = orm.mapper(IasiSurvey, self._t_survey)
        
        self._session_class = orm.sessionmaker(bind = self._engine, autoflush = True, transactional = True)

    def session(self):
        """return a transaction session allowing us to query and save objects instead of rows
        object proxies to database table entries can be manipulated and then committed back with the database session
        e.g.
        S=db.session()
        (q,)=S.query(Granule).filter(Granule.name=='IASI_xxx_1C_M02_20080216124758Z_20080216125054Z_N_O_20080216143228Z__20080216143824')
        print q.replicas
        print q.keywords
        print q.name
        """
        # FUTURE: review sqlalchemy notes on whether we really need this anymore
        return self._session_class()
                
    def __init__(self, db_url=None, create=False):
        if not db_url: db_url = DATABASE_URL
        self.connect(db_url,create=create)
        self.create_mappings()
        
    def contains_name(self,name):
        return len( self._t_simple.select(self._t_simple.c.name==name).execute().fetchall() )>0
        
    def insert_simple_entry(self, **nfo):
        """
        insert a new entry with zero or more URLs of access
        .insert_entry(  name = 'IASI_____________', 
                        instrument='iasi', start=datetime(), end = datetime(),
                        perimeter=( (lon0,lat0), (lon1,lat1), (lon2,lat2), (lon3,lat3) ),
                        keywords=['cloud', 'tc4', 'cart'], state='alpha', 
                        description='processed IASI retrievals using uwphysret 20071214' )
        returns UUID object which can be used to build replica rows and scanline rows
        """
        uu = uuid.uuid1()
        LOG.info('creating entry %s' % uu)
        LOG.debug('info for it is %s' % nfo)
        # ensure that we have all the necessary metadata in the nfo parameter dictionary
        keze = ['name', 'instrument', 'perimeter', 'start_time', 'end_time', 'state', 'description']
        simple = dict( (k,nfo[k]) for k in keze if k in nfo )
        # calculate span as a line segment in unix epoch from start to end
        if 'start_time' in nfo:
            estart,eend = iasi.epoch_from_datetime( [nfo['start_time'], nfo['end_time']] )
        else:
            estart = eend = float('nan')
        span = ( (estart,0), (eend,0) )
        # insert the simple metadata
        self._t_simple.insert().execute(uuid = uu, span = span, **simple) 
        return uu
        
    def insert_replica_entries(self, uu, url_list, **nfo):
        """associate one or more replicas with a resource
        """
        if not isinstance(uu,uuid.UUID): uu = uuid.UUID(uu)
        # insert all the links we can access through
        (name,) = select([self._t_simple.c.name], self._t_simple.c.uuid==uu).execute().fetchall()
        for url in url_list:
            LOG.info('associating %s with %s' % (url,name))
            self._t_replica.insert().execute( uuid = uu,
                                        content_url = url,
                                        quicklook_url = nfo.get('quicklook_url',''),  # FUTURE: make this do something useful
                                        incept = datetime.utcnow(),
                                        seen = datetime.utcnow(),
                                        state = nfo.get('state',''),
                                        expiry=nfo.get('expiry',None),
                                        protocol=nfo.get('protocol',"")
                                        )

    def insert_keywords(self, uu, kival):
        """insert an iterable set of keywords with optional values, from a list or dictionary
        """
        if not isinstance(uu,uuid.UUID): uu = uuid.UUID(uu)
        isdict = hasattr(kival,'get')
        (name,) = select([self._t_simple.c.name],self._t_simple.c.uuid==uu).execute().fetchall()
        LOG.debug("updating keywords table for %s (%s)" % (name,uu))
        for key in kival:
            doesexist = self._t_keywords.select( (self._t_keywords.c.uuid==uu) & (self._t_keywords.c.key==key) ).execute().fetchall()
            value = str(kival.get(key)) if isdict else None
            if doesexist:
                LOG.debug("updating %s = '%s' -> '%s'" % (key,doesexist.value,value))
                self._t_keywords.update( (self._t_keywords.c.uuid==uu) & (self._t_keywords.c.key==key) ).execute(value = value)
            else:
                LOG.debug("adding %s = '%s'" % (key,value))
                self._t_keywords.insert().execute( uuid = uu, key = key, value = value )

    def delete_uuid(self,uu):
        """purge a granule from the database, given its uuid
        note: assumes that uuid column is set on fields, which is not necessarily true for old data
        """
        if not isinstance(uu,uuid.UUID): uu = uuid.UUID(uu)
        self._t_for.delete(self._t_for.c.uuid==uu).execute()
        self._t_keywords.delete(self._t_keywords.c.uuid==uu).execute()
        self._t_scan.delete(self._t_scan.c.uuid==uu).execute()
        self._t_replica.delete(self._t_replica.c.uuid==uu).execute()
        self._t_simple.delete(self._t_simple.c.uuid==uu).execute()

    def delete_scanlines(self, uu):
        """remove scanlines and fields of regard for a given granule uuid
        note: note all field rows have uuid column set
        this is generally best used before doing a re-insert_scanlines_from_generator() 
        """
        if not isinstance(uu,uuid.UUID): uu = uuid.UUID(uu)
        self._t_for.delete(self._t_for.c.uuid==uu).execute()
        self._t_scan.delete(self._t_scan.c.uuid==uu).execute()
                                                        
    def maia_columns(self, maia_info, scandex, fordex, detdex):
        """ return dictionary of columns which should go into the field of regard table, given data from maia
        """
        nfo = maia_info.get( (scandex,fordex,detdex), {} )
        # tersect = set(nfo.keys()) & set()
        zult = dict((k,nfo[k]) for k in self.FIELD_OF_VIEW_KEYS if k in nfo)
        LOG.debug('maia data columns found: %s' % str(zult))
        return zult
        
    def insert_scanlines_from_generator(self, uu, scans, maia_info = { }, include_field_of_regard=True, dry_run=False, **nfo):
        """ insert scan-line metadata for a resource from a scan-line generator returning a sequence of dictionaries
        additional keyword parameters contains additional metadata to assign to each scan (e.g. shopping list keywords)
        maia_info is a dictionary of { (scanline,field,detector): {column-data}, ... } and is optional
        see iasi_tools.generate_scanline_metadata
        """        
        if not isinstance(uu,uuid.UUID): uu = uuid.UUID(uu)
        # insert all the links we can access through
        # (name,) = select([self._t_simple.c.name],self._t_simple.c.uuid==uu).execute().fetchall()
        LOG.info('inserting scan lines for %s' % str(uu))
        for scan in scans:
            if scan['start_time'].year <= 2001 or scan['end_time'].year <= 2001:
                LOG.warning('scan %s has bad time! skipping it' % scan['scan_index'])
                continue
            scan_time = datetime(*scan['start_time'].utctimetuple()[0:6])  # truncate microseconds
            # calculate span as a line segment in unix epoch from start to end
            estart,eend = iasi.epoch_from_datetime( [scan['start_time'], scan['end_time']] )
            span = ( (estart,0), (eend,0) )
            sert = dict( scan )
            sert.update(nfo)
            LOG.debug( "scan metadata for %s: %s" % (uu, sert) )
            if not dry_run:
                self._t_scan.insert().execute( uuid = uu, scan_time = scan_time, span = span, **sert )
            else:
                LOG.debug('inserting scan line at %s' % str(scan_time))
            # note: this now handles fields of regard or fields of view depending on the scanline generator settings
            # detector number comes through as None if we're generating fields of regard
            if include_field_of_regard: # insert field of regard navigation data 
                LOG.debug("inserting %s field metadata entries" % len(scan['for_time']))

                # merge cloud info into database insert
                # bighonkinginsert = [ dict(for_time=when, for_center=where, for_index_within_scan=fordex, for_detector_number=detdex ) # FIXME , **self.maia_columns(maia_info, scan['scan_index'], fordex, detdex) )
                bighonkinginsert = [ dict(uuid=uu, for_time=when, for_center=where, for_index_within_scan=fordex, for_detector_number=detdex, **self.maia_columns(maia_info, scan['scan_index'], fordex, detdex) )
                                     for when,where,fordex,detdex in zip(scan['for_time'], scan['for_center'], scan['field_index'], scan['detector_index'])]
                if not dry_run:
                    # for honk in bighonkinginsert: self._t_for.insert().execute( honk ) #*bighonkinginsert ) FIXME
                    self._t_for.insert().execute( *bighonkinginsert ) 
                else:
                    from pprint import pformat
                    LOG.debug("big honking insert: %s" % pformat(bighonkinginsert))
                                        
    def uuid_for_name(self, name):
        ((result,),) = select([self._t_simple.c.uuid], self._t_simple.c.name==name).execute().fetchall()
        return result

    def name_for_uuid(self, uu):
        if not isinstance(uu,uuid.UUID): uu = uuid.UUID(uu)
        ((result,),)= select([self._t_simple.c.name], self._t_simple.c.uuid==uu).execute().fetchall()
        return result
                                                
    def touch(self, url):
        "update the last-seen time of a replica represented by a URL"
        self._t_replica.update(self._t_replica.c.content_url == url).execute(seen = datetime(*gmtime()[0:6]))
    
    def get_inventory(self):
        """ return set of names in the cache as a dictionary of { name : uuid }
        """
        return dict( select( [self._t_simple.c.name, self._t_simple.c.uuid] ).execute().fetchall() )
                
    # where clauses which we bolt into sqlalchemy as text objects
       
    @staticmethod         
    def _where_box_holds_point(box_tuple, point_column):
        """where clause which checks if a point column falls inside a bounding box
        """
        return text( "'%s'::box::polygon ~ %s" % (str(box_tuple), str(point_column)) )

    @staticmethod         
    def _where_polygon_holds_point(box_tuple, point_column):
        """where clause which checks if a point column falls inside a polygon
        """
        return text( "'%s'::polygon ~ %s" % (str(box_tuple), str(point_column)) )
        
    @staticmethod         
    def _where_box_intersects_polygon(box_tuple, polygon_column):
        """where clause which checks if a perimeter polygon intersects the provided box
        """
        return text( "'%s'::box::polygon && %s" % (str(box_tuple), str(polygon_column)) )
        
    @staticmethod         
    def _where_polygon_intersects_polygon(polygon_tuple, polygon_column):
        """where clause which checks if a perimeter polygon intersects the provided box
        """
        return text( "'%s'::polygon && %s" % (str(polygon_tuple), str(polygon_column)) )

    @staticmethod
    def _where_lsegs_overlap(segment, lseg_column):
        # NOTE: watch for bug in 8.2.x where parallel intersecting lsegs (lseg overlap) get false negative
        # our use of lsegs to represent time spans depends on ?# being implemented correctly
        # see http://archives.postgresql.org/pgsql-bugs/2008-03/msg00099.php
        return text( "'%s'::lseg ?# %s" % (str(segment), str(lseg_column)) )

    @staticmethod
    def _where_time_segments_overlap( (a,b), (c,d) ):
        return not_( (a>d) | (c>b) )

    # search types

    def scans_in_box(self, box):
        """ return iterable list of scan metadata intersecting a longitude-latitude box represented as ( (lon,lat), (lon,lat) )
        """
        return self._t_scan.select( self._where_box_intersects_polygon(box, self._t_scan.c.perimeter) ).execute().fetchall()

    def scans_in_polygon(self, poly):
        """ return iterable list of scan metadata intersecting a longitude-latitude box represented as ( (lon,lat), (lon,lat) )
        """
        return self._t_scan.select( self._where_polygon_intersects_polygon(poly, self._t_scan.c.perimeter) ).execute().fetchall()
        
    def urls_for_uuid(self, uu=None, name=None):
        """ return a list of urls where a given uuid can be accessed
        """
        if uu is None and name is not None:
            uu = self.uuid_for_name(name)
        if not isinstance(uu,uuid.UUID): uu = uuid.UUID(uu)
        return [url for (url,) in select( [self._t_replica.c.content_url], self._t_replica.c.uuid == uu ).execute().fetchall()]
        
    def granules_without_replicas(self):
        """ return dictionary of uuid:name pairs having no replicas
        """
        names = dict( (x.uuid, x.name) for x in self._t_simple.select().execute().fetchall())
        replicas = dict( (x.uuid, x.content_url) for x in self._t_replica.select().execute().fetchall())
        noreplicas = set( replicas.keys() ) ^ set( names.keys() )
        return dict( (uu,names[uu]) for uu in noreplicas )
        
    def fields_without_granules(self):
        return self._t_for.select(self._t_for.c.uuid==None).execute()
        
    def granules_with_too_few_scans(self, ok=5):
        """ return dictionary of uuid:name pairs having too few scans indexed
        e.g. 
        for (uu,name) in db.granules_with_too_few_scans().items(): 
            print "marking %s bad" % name
            db.insert_keywords( uu, ['bad'] )
        """
        names = dict( (uuid.UUID(x.uuid), x.name) for x in self._t_simple.select().execute().fetchall())
        count = dict( (k, 0) for k in names.keys() )
        for x in self._t_scan.select().execute().fetchall():
            count[uuid.UUID(x.uuid)] += 1
        zult = { }
        for k in count.keys():
            if count[k] >= ok:
                 del names[k]                
        return names
        
    def _search_fields_of_view(self, bounding_box=None, time_range=None, bounding_polygon=None, for_list=None, string_criteria=None):
        """ yield granule-uuid, [ (field-id, scanline-index, field-index, detector-index), ... ] for search results
        scan-list is a list of (scan-index, [list-of-fields-of-regard-indices])
        bounding_box is longitude-latitude box represented as ( (lon,lat), (lon,lat) )
        time_range is ( datetime1, datetime2 ) object tuple representing datetime1 <= T < datetime2
        bounding_polygon is a series of points in a tuple representing a closed polygon to intersect with
        for_list is a list of field-of-regard numbers 0..29
        string_criteria is a list of expressions to evaluate to where clauses, e.g. cloudy_percent<50
        these are principally MAIA columns
        """
        # based on input parms, set up criteria for fields and granules to apply server-side
        griteria = []   # criteria to apply to granules
        friteria = []   # criteria to apply to fields
        for num_expr in (string_criteria or []):
            LOG.info("constructing where-expr from '%s'" % num_expr)
            context = dict((k,getattr(self._t_for.c,k)) for k in self.FIELD_OF_VIEW_KEYS)
            context['detector']=self._t_for.c.for_detector_number
            LOG.debug("context is %s" % str(context))
            where_expr=eval( num_expr, context, context )
            friteria.append(where_expr)
            LOG.debug('%s: %s' % (type(where_expr), str(where_expr)))
        if time_range is not None:
            stime, etime = time_range
            estart,eend = iasi.epoch_from_datetime( time_range )
            lseg = ((estart,0),(eend,0))
            friteria.append( (self._t_for.c.for_time >= stime) & (self._t_for.c.for_time < etime) )
            griteria.append( self._where_time_segments_overlap( (self._t_simple.c.start_time, self._t_simple.c.end_time),
                                        (stime,etime)) )
        if bounding_polygon is not None:
            friteria.append( self._where_polygon_holds_point(bounding_polygon, self._t_for.c.for_center) )
            griteria.append( self._where_polygon_intersects_polygon(bounding_polygon, self._t_simple.c.perimeter))
        elif bounding_box is not None: 
            friteria.append( self._where_box_holds_point(bounding_box, self._t_for.c.for_center) )
            griteria.append( self._where_box_intersects_polygon(bounding_box, self._t_simple.c.perimeter) )
        if for_list is not None:
            friteria.append( self._t_for.c.for_index_within_scan.in_(for_list) )
        if not friteria: 
            return
        # resolve granules of interest
        granule_where_expr = and_(*griteria) if len(griteria)>1 else griteria[0]
        lect = select( [self._t_simple.c.uuid, self._t_simple.c.start_time, self._t_simple.c.end_time, self._t_simple.c.name], granule_where_expr)
        # print lect
        LOG.debug('granule select expression: %s' % str(lect))
        granules = lect.execute().fetchall()
        LOG.info('found %d granules with fields possibly fitting search criteria:' % len(granules))
        for _,_,_,name in granules:
            LOG.info(name)
        # search individual granules' fields
        for uu,stime,etime,name in granules:
            LOG.info("checking %s..." % name)
            if stime is None or etime is None: # we've had problems with bad times in granules
                LOG.debug('this granule is bad! skipping')
                continue
            criteria = list(friteria)
            criteria.insert( 0, self._t_scan.c.uuid == uu )
            criteria.append( (self._t_for.c.for_time >= stime) & (self._t_for.c.for_time < etime) )
            criteria.append( (self._t_for.c.for_time >= self._t_scan.c.start_time) & (self._t_for.c.for_time < self._t_scan.c.end_time) )
            criteria.insert( 0, self._t_for.c.uuid == uu )  # match field of view to granule
            where_expr = and_(*criteria)
            sully = select( [self._t_for.c.id, self._t_scan.c.scan_index, self._t_for.c.for_index_within_scan, self._t_for.c.for_detector_number], where_expr )
            LOG.debug(str(sully))
            scanfors = sully.execute().fetchall()
            LOG.debug("found %d fields-of-view in granule at %s matching search criteria" % (len(scanfors),str(stime)))
            if len(scanfors)==0: continue
            yield uu, scanfors
            
    def search_fields_of_regard(self, bounding_box=None, time_range=None, bounding_polygon=None, for_list=None, string_criteria=None):
        """generate ( granule-uuid, { scanline: { field: set(detectors), ... }, ...} ) for a given bounding_box and/or time_range
        scan-list is a list of (scan-index, [list-of-fields-of-regard-indices])
        bounding_box is longitude-latitude box represented as ( (lon,lat), (lon,lat) )
        time_range is ( datetime1, datetime2 ) object tuple representing datetime1 <= T < datetime2
        bounding_polygon is a series of points in a tuple representing a closed polygon to intersect with
        for_list is a list of field-of-regard numbers 0..29
        string_criteria is a list of expressions to evaluate to where clauses, e.g. cloudy_percent<50
        these are principally MAIA columns
        """
        for uu,scanfors in self._search_fields_of_view(bounding_box, time_range, bounding_polygon, for_list, string_criteria):
            Q = dict()
            for field_id,scandex,fordex,detdex in scanfors: 
                fields = Q.setdefault(scandex,dict())
                detectors = fields.setdefault(fordex,set())
                if detdex is None: detectors |= set(range(4))
                else: detectors.add(detdex)
            yield uu, Q
            
    def search_fields_as_objects(self, bounding_box=None, time_range=None, bounding_polygon=None, for_list=None, string_criteria=None):
        """ generate IasiField objects for a search
        scan-list is a list of (scan-index, [list-of-fields-of-regard-indices])
        bounding_box is longitude-latitude box represented as ( (lon,lat), (lon,lat) )
        time_range is ( datetime1, datetime2 ) object tuple representing datetime1 <= T < datetime2
        bounding_polygon is a series of points in a tuple representing a closed polygon to intersect with
        for_list is a list of field-of-regard numbers 0..29
        string_criteria is a list of expressions to evaluate to where clauses, e.g. cloudy_percent<50
        these are principally MAIA columns
        """
        S = self.session()
        for uu,scanfors in self._search_fields_of_view(bounding_box, time_range, bounding_polygon, for_list, string_criteria):
            field_ids = [x[0] for x in scanfors]
            for fob in S.query(IasiField).filter(self._t_for.c.id.in_(field_ids)):
                yield fob                    

    def _search_granule(self, names=None, time_range=None, keywords_all_any_not=None, bounding_box=None, sun_elevation_angle_range=None, cloud_coverage_range=None, over_sea=None):
        pass
        
    def _search_scan(self, names=None, time_range=None, keywords_all_any_not=None, bounding_box=None, sun_elevation_angle_range=None, cloud_coverage_range=None, over_sea=None):
        pass

    def _search_field(self, names=None, time_range=None, keywords_all_any_not=None, bounding_box=None, sun_elevation_angle_range=None, cloud_coverage_range=None, over_sea=None):
            
        pass

    def search(self, names=None, time_range=None, keywords_all_any_not=None, bounding_box=None, sun_elevation_angle_range=None, cloud_coverage_range=None, over_sea=None, resolution=None):
        """ general query-by-example routine
        Yields a series of tuples in the form ( uuid, [ (scan_index, [for_index, ...]), ... ] )
        In the case of all-scans or all-fields, None will be used to indicate no constraint
            e.g. ( uuid, None ) or (uuid, [ (scan_index, None), (scan_index, None) ])
        
        Criteria can remain 'None' for don't-cares:
            names: list of granule names to include
            time_range: time range constraint, (start,end) as datetime objects
            keywords_all_any_not: 3-tuple: ( [all-list], [any-list], [not-list] ) 
                each list is a list of keyword strings and can be empty
            bounding_box: ((lon,lat),(lon,lat)) tuple
            sun_elevation_angle_range: (min,max) tuple, angles in degrees
            cloud_coverage: (min,max) cloud coverage as 0..1 values
            over_sea: True,False
            resolution: None, 'granule', 'scan', 'field', defaults to 'field'
        """
        if time_range is None and bounding_box is None and (resolution is None or resolution=='field'): 
            resolution = 'scan'
        # FIXME
        raise NotImplementedError
        
    def update_field(self, uu, (scanline,field,detector), new_data):
        """update the fields in the new_data dictionary into the corresponding field
        """
        self._t_for.update( self._t_for.c.uuid==uu and 
                            self._t_scan.c.uuid==uu and
                            self._t_scan.c.start_time <= self._t_for.c.for_time and
                            self._t_scan.c.for_time < self._t_scan.c.end_time and
                            self._t_scan.c.scan_index==scanline and 
                            self._t_for.c.for_index_within_scan==field and
                            self._t_for.c.for_detector_number==detector ).execute( **new_data )

    def search_scans(self, time_range):
        """ generate ( uuid, [ (scan_index, [for_index, ...]), ... ] ) for a given time range
        scan-list is a list of (scan-index, [list-of-fields-of-regard-indices])
        time_range is ( datetime1, datetime2 ) object tuple representing datetime1 <= T < datetime2
        """
        LOG.debug('searching scans for time range %s-%s' % tuple(time_range))
        criteria = []
        stime, etime = time_range
        criteria.append( (self._t_scan.c.start_time >= stime) & (self._t_scan.c.end_time < etime) )
        if not criteria: 
            return
        where_expr = and_(*criteria) if len(criteria)>1 else criteria[0]
        scans = select( [self._t_scan.c.uuid, self._t_scan.c.scan_index], where_expr ).execute().fetchall()
        LOG.info("found %d scans matching search criteria" % len(scans))
        D = defaultdict(set)
        for uu,scandex in scans:
            # find scans that the fields belong to
            D[uu].add( scandex )
        for uu,scanset in D.items():
            scanlist = list(scanset)
            scanlist.sort()
            # FUTURE: resolve individual fields within scans that fit the time range - right now we return all 30
            yield uu, [ (scandex,range(0,30)) for scandex in scanlist ]

    def granule_for_uuid(self, uu):
        if not isinstance(uu,uuid.UUID): uu = uuid.UUID(uu)
        (gran,) = self._t_simple.select( self._t_simple.c.uuid==uu ).execute().fetchall()
        return gran
        
    #
    # survey table support
    #
    
    def update_survey(self, name, content_uri):
        """ update survey table with URI of survey pickle file
        FUTURE: put this in a transaction
        """
        nfo = dict(name=name, content_uri=content_uri, creation_time=datetime.utcnow(), ingest_time=None)
        
        lol = self._t_survey.select(self._t_survey.c.name == name).execute().fetchall()
        assert(len(lol) <= 1)
        if len(lol)==1:
            del nfo['name']
            self._t_survey.update(id = lol[0].id).execute(**nfo)
        else:
            self._t_survey.insert().execute(**nfo)

    def survey_entry_for_name(self,name):
        lol = self._t_survey.select(self._t_survey.c.name == name).execute().fetchall()
        assert(len(lol)<=1)
        if not lol: return None
        return lol[0]
        
    def pick_unfinished_survey_entry(self):
        """check out an unfinished entry from the survey table, as a transaction
        returns None if transaction failed
        use abandon_unfinished_survey_entry() to return it if it doesn't get completed
        """
        trans = self._connection.begin()
        try:
            eet = self._t_survey.select(self._t_survey.c.ingest_time == None).execute().fetchone()
            self._t_survey.update((self._t_survey.c.id==eet.id) & (self._t_survey.c.ingest_time==None)).execute(ingest_time=datetime.utcnow())
            trans.commit()
            return eet
        except:
            trans.rollback()
            return None
        
    def abandon_unfinished_survey_entry(self,entry):
        "return an unfinished survey entry to the to-be-ingested list"
        self._t_survey.update((self._t_survey.c.id==entry.id) & (self._t_survey.c.ingest_time!=None)).execute(ingest_time=None)

    #
    # keywords table
    #

    def keywords_for_uuid(self, uu=None, name=None, include_all=False):
        """ return dictionary of keywords associated with a given uuid
        """
        if uu is None and name is not None:
            uu = self.uuid_for_name(name)
        if not isinstance(uu,uuid.UUID): uu = uuid.UUID(uu)
        lol = select( [self._t_keywords.c.key, self._t_keywords.c.value], 
                                self._t_keywords.c.uuid==uu ).execute().fetchall()   
        if include_all: # FUTURE: how to do this better? this is kinda crufty
            all_keys = select( [self._t_keywords.c.key] ).execute().fetchall()
            D = dict( (k,False) for (k,) in set(all_keys) )
        else:
            # use a defaultdict with a default of False so that keyword expressions don't bjork (Does not work with eval()! sry)
            D = defaultdict(bool)
        # FUTURE: use pickle or json for values in table
        D.update( dict( (k,True if v=='' else v) for k,v in lol ) )
        return D
    
    def all_keywords(self):
        """ return { keyword: set-of-uuids } dictionary of all keywords in the database
        """
        D = defaultdict(set)
        lal = self._t_keywords.select().execute().fetchall()
        for each in lal: 
            D[each.key].add(each.uuid)
        return D
        
    # def iasi_backfill_field_uuids(self, dry_run=False):
    #     """ fill uuid fields for fields-of-regard table; should not leave anything behind
    #     """    
    #     # FIXME: this doesn't work right yet
    #     raise NotImplementedError
    #     # for each granule
    #     for granule in self._t_simple.select(self._t_simple.c.instrument=='iasi').execute():
    #         # get the times of the fields which fall inside the granule
    #         fields = select( [self._t_for.c.for_time],
    #                                 (granule.start_time <= self._t_for.c.for_time) & 
    #                                 (self._t_for.c.for_time <= granule.end_time)
    #                                 ).execute().fetchall()
    #         if dry_run:
    #             print granule.name, len(fields), fields[:20]
    #         else:
    #             print '\n\n' + granule.name
    #             for fieldtime in fields:
    #                 # FUTURE: buffer up these updates?
    #                 self._t_for.update(self._t_for.c.for_time==fieldtime).execute(uuid=granule.uuid)
    #                 print '.', 
    #                 sys.stdout.flush()    
    #     print "orphan list:"
    #     print select( [self._t_for.c.for_time], self._t_for.c.uuid==None ).execute().fetchall()

    def test_insert(self):
        uu = uuid.UUID('0' * 32)
        self._t_simple.insert().execute( uuid = uu, 
                                        name = 'test_info',
                                        instrument = 'None',
                                        perimeter = ( (0,1), (0.5,1.5), (0,2), (1,2), (0.8,1.4), (1,1) ),
                                        # bounds = ( (0,1), (0,2), (1,2), (1,1) ),
                                        start = datetime( 1980, 12, 25, 12, 0, 0 ),
                                        end = datetime (1980, 12, 25, 13, 0, 0 ),
                                        span = ( (0,0), (1.4, 0) ),
                                        keywords = "my,dog,has,fleas",
                                        state = "test",
                                        description = "test insert by $Id: iasi_cache.py,v 1.124 2008/08/18 21:38:21 rayg Exp $"
                                        )
        self._t_replica.insert().execute( uuid = uu, 
                                        content_url = 'http://www.ssec.wisc.edu', 
                                        protocol = 'http',
                                        quicklook_url = 'http://www.ssec.wisc.edu',
                                        incept = datetime(*gmtime()[0:6]),
                                        expiry = datetime(*gmtime()[0:6]) + timedelta(4), # 4 days later
                                        seen = datetime(*gmtime()[0:6])
                                        )
    def test_remove(self):
        uu = uuid.UUID('0' * 32)
        self._t_replica.delete(self._t_replica.c.uuid==uu).execute()
        self._t_simple.delete(self._t_simple.c.uuid==uu).execute()


