#! /usr/bin/env python
""" $Id: iasi_tools.py,v 1.46 2009/03/19 23:24:23 rayg Exp $
Toolbox of useful IASI functions by way of EUGENE file access
"""
import sys
import os
import re
import calendar
import logging
from datetime import datetime
from collections import defaultdict
from scipy.fftpack import rfft,irfft
from scipy.interpolate.interpolate import interp1d
from pprint import pformat
from numpy import exp,log,array,arange,empty,float32,float64,sin,linspace,concatenate,repeat,reshape,rollaxis
from math import pi

CVSID="$Id: iasi_tools.py,v 1.46 2009/03/19 23:24:23 rayg Exp $"

LOG = logging.getLogger('iasi_tools')

# format string for printing out information about an iasi_record
IASI_RECORD_FMT = """<IASI Sounding record %(record_name)s scan %(scan_number)d detector %(detector_number)d time %(time)s latitude %(latitude)f longitude %(longitude)s>"""

class iasi_record(object):
    """ Simple data holder for IASI sounder records, i.e. a struct.
    Used as a return data container by generate_ functions.
    """
    record_name = None
    scan_number = None
    detector_number = None
    
    spectralFreq = None # array
    radiances = None # array
    time = None     # datetime.datetime
    epoch = None    # double precision, unix epoch time
    latitude = None # float
    longitude = None # float
    scanAng = None # float
    
    metop_zenith_angle = None # degrees
    metop_aziumth_angle = None
    sun_zenith_angle = None
    sun_azimuth_angle = None
    
    landFrac = None # float
    elevation = None # float
    
    def __repr__(self):
        return IASI_RECORD_FMT % vars(self)
        
    def __str__(self):
        return pformat(vars(self))
        

class iasi_profile(object):
    """ Simple data holder for IASI retrieval records.
    Not currently used. 
    """
    record_name = None
    scan_number = None
    detector_number = None
    
    


def epoch_from_datetime(x):
    """Convert datetime.datetime object to double precision UNIX epoch time in seconds.
    """
    if isinstance(x,datetime):
        return calendar.timegm( x.utctimetuple() )
    else:
        return [float(calendar.timegm( q.utctimetuple() )) + float(q.microsecond)/1e6 for q in x]

class iasi_image(object):
    """a container object for map data
    """
    latitude = None
    longitude = None
    

CLOUD_MASK_PRODUCTS = [ 'GCcsRadAnalNbClass', 'GCcsRadAnalWgt', 'GCcsRadAnalY', 'GCcsRadAnalZ', 'GCcsRadAnalMean', 'GCcsRadAnalStd' ]

def scale_scanline(prod, spectra=None, factor_table = None):
    """Apply scaling factors for IASI L1B spectra (spectra[spectrumnumber][wavenumber] indexed).
    Returns factor_table which can be used to speed up subsequent calls.
    Output spectra are in mW/m2 sr cm-1
    """
    if not factor_table: 
        scaling = prod.get('giadr-scalefactors')
        nscales = scaling.IDefScaleSondNbScale
        offset = scaling.IDefScaleSondNsfirst[0]
        factor_table = zip( range(nscales),
                            array(scaling.IDefScaleSondNsfirst)-offset,
                            array(scaling.IDefScaleSondNslast)-offset+1,
                            array(scaling.IDefScaleSondScaleFactor)  
                            )
    if spectra!=None:
        for _,start,end,factor in factor_table: 
            spectra[:,start:end] *= 10.**(-factor+5)  # UW scaling preferred has a 1e5 difference : nets us mW/m2 sr cm-1
    return factor_table

def sounder_wavenumber_interval(prod):
    """ Return (vmin, dv, npts) wavenumber interval tuple in cm-1
    """
    IDefNsfirst1b = prod.get("mdr-1c[0].IDefNsfirst1b")
    IDefSpectDWn1b = prod.get("mdr-1c[0].IDefSpectDWn1b")
    IDefNslast1b = prod.get("mdr-1c[0].IDefNslast1b")
    delta = IDefSpectDWn1b / 100. # convert to cm-1
    return delta * (IDefNsfirst1b-1), delta, IDefNslast1b - IDefNsfirst1b + 1

def sounder_wavenumbers(prod):
    """Return sounder wavenumber scale as an array and a delta
    for sample #k: wavenumber(k) = IDefSpectDWn1b * (IDefNsfirst1b + k -2)
    FUTURE: account for any variation that probably doesn't exist between records
    Currently, only the first record's values are being returned. 
    Also note that the wavenumber array is shorter than the 8700-wide spectrum
    """
    IDefNsfirst1b = prod.get("mdr-1c[0].IDefNsfirst1b")
    IDefSpectDWn1b = prod.get("mdr-1c[0].IDefSpectDWn1b")
    IDefNslast1b = prod.get("mdr-1c[0].IDefNslast1b")
    delta = IDefSpectDWn1b / 100. # convert to cm-1
    return arange(IDefNsfirst1b-1,IDefNslast1b,dtype=float64) * delta, delta
    
def cosine_rolloff(vin, rin, vb):
    """
    Apply cosine roll-offs to edges of an input spectrum
    
    Inputs:
     vin,rin: input wavenumbers,radiances
     vb[0,1]: min and max boundaries where roll-off function is applied
     
    Outputs:
     rout: modified radiances
     rout[-1] is 0.0
     rout ramps to rin@vb[0] on a sine wave
     rout ramps from rin@vb[1] on a sine wave
     rout[0] is 0.0

    The input spectrum for v < v_rolloff(1) and for v > v_rolloff(2)
    is replaced with the appropriate cosine function.

    from rolloff.m by DCTobin 2-12-99
    
    FUTURE: promote this to a spectral toolbox module
    
    Testing:
    >>> from numpy import ones, arange
    >>> x = arange(0,100,dtype='d')
    >>> y = ones(100,dtype='d')
    >>> z = cosine_rolloff(x,y,[20.,60.])
    """
    rout = array(rin)
    
    # find the index where vb[0] would be inserted
    left = vin.searchsorted(vb[0], 'left')
    # find the ramp's ultimate maximum value without going outside of bounds
    endval = rin[min(left, len(rin) - 1)]
    # build a cosine ramp of the appropriate width
    ramp = linspace(pi/-2., pi/2., left)
    # multiply through while replacing elements
    rout[0:left] = (sin(ramp) / 2. + 0.5) * endval

    # second verse! go in reverse!
    right = vin.searchsorted(vb[1], 'right')
    width = len(rout) - right
    endval = rin[max(0, right-1)]
    ramp = linspace(pi/2., pi/-2., width)
    rout[right:] = (sin(ramp) / 2. + 0.5) * endval
    
    return rout
    
def sounder_interpolating_dv(panel_dvs_list, dv_final):
    """Compute dv interval for interpolating monochromatic LBL panel data to.
    dv_final can be found with sounder_wavenumbers()
    From an algorithm sketch by BobK 200710
    FUTURE: promote to spectral toolbox module
    """
    # Find the minimum dv_lbl contained among the panels contained in the LBLRTM TAPE12 file(s) to be used. There can be multiple TAPE12 files. 
    dv_lbl = min(panel_dvs_list)
    
    # determine the desired dv_final of the output of the FFT (user input based upon the instrument sampling that is being modeled). 
    # Note this is often computed as some power of 2 times a fundamental laser frequency for FTS systems. 
    # if dv_final is None:
    #     dv_final = (sampling_factor * v_laser) / (sample_count * reduction_factor)
    
    #  Find the power of 2 for which  dv_interp = dv_final / 2^M < dv_lbl,  
    # i.e. choose M such that a submultiple of the final output is less than the LBLRTM spacing. 
    # (This ensures that there is no loss of information in the linear interpolation between LBLRTM points. (Any lower sampling can lead to large interpolation errors!) 
    # dv_final / 2**m < dv_lbl
    # let dv_lbl = dv_final * 2**-m
    # log2(dv_lbl) = log2(dv_final) - m
    from numpy import log2, ceil
    m = log2(dv_final) - log2(dv_lbl)
    M = ceil(m)
    dv_interp = dv_final / 2**M 
    
    return dv_interp
    
def sounder_scan_monochromatic_radiances(v_mono, r_mono, zflag = False, v_rolloff = [None,None], apodize = True):
    """
    apply IASI L1-C instrument function to an input monochromatic radiance spectrum
    from iasi_1c.m by DCTobin
    
    Input variables:
     v_mono are the input monochromatic wavenumbers.  Should be very highly sampled.
     r_mono are the input monochromatic radiances in units of W/(m^2 ster cm^-1)
     zflag determines if the apodized interferogram is truncated or zero filled.
    	if zflag == True, the interferogram points past MOPD are retained, but zero 
     	filled and the output spectrum has a finer point spacing.  
     	if zflag == False (the default), the interferogram is trucated past MOPD 
    	and the resulting output spectrum is reported at the 0.25 1/cm spacing.
     v_rolloff is a two element array defining where the boundaries
           (in 1/cm) of where the rolloff function is applied.  See rolloff.m for info.
     apodize is a boolean (default True) which toggles whether to apply gaussian apodization or not
     returns ( resampled-radiances, resampled-wavenumber )
    """    
    # default rolloff range is 5% extension on either side
    # FIXME: should this be done after v_mono truncation?
    # note: the assign-to-list syntax is used to avoid possible reassignment of default parameter
    if v_rolloff[0] is None:
        v_rolloff = [0.05*(v_mono[-1]-v_mono[0]) + v_mono[0], v_rolloff[1]]
    if v_rolloff[1] is None:
        v_rolloff = [v_rolloff[0], -0.05*(v_mono[-1]-v_mono[0]) + v_mono[-1]]
        
    # force an odd number of points
    npts = min(len(r_mono),len(v_mono)) 
    npts -= (npts % 2) ^ 1 
    
    # add cosine rolloff and truncate 
    LOG.info('cosine rolloff')
    v_mono = v_mono[0:npts]
    r_mono = cosine_rolloff(v_mono,r_mono[0:npts],v_rolloff);

    # rearrange as symmetric real spectrum and perform fft
    def _radfold(x): return array( [x[-1:0:-1], x[0:-1]] ).reshape(2 * (len(x) - 1))
    rad2 = _radfold(r_mono)
    LOG.info('forward FFT')
    yfft = rfft(rad2,overwrite_x=True)
    del rad2
    
    # Compute corresponding OPD scale
    dx = 0.5 / (v_mono[-1] - v_mono[0])
    mopd = npts * dx
    opd = linspace(0,mopd,npts)
    def _opdfold(x): return array( [x[0:-1], -x[-1:0:-1]] ).reshape(2 * (len(x) - 1))
    opd2 = _opdfold(opd)
    del opd

    # Apply IASI L1C truncated Gaussian apodization
    if apodize: 
        LOG.info('apodizing')
        fwhm = 0.5;
        MaxOPD = 1.96856;
        a = 4. * log(2.) / (fwhm * fwhm);
        pisq = pi * pi
        gresp2 = exp(-pisq * opd2 ** 2 / a)
        gresp2[ abs(opd2) > MaxOPD ] = 0.0
        yfft *= gresp2
    
    # Compute correpsonding spectrum
    LOG.info('inverse FFT')
    spec = irfft(yfft,overwrite_x=True)
    del yfft
    spec = spec[int(len(spec) / 2):]
    
    # truncate to the rolloff area by finding leftmost and rightmost insertion indices
    if zflag: 
        dexmin = v_mono.searchsorted( v_rolloff[0], 'left' )
        dexmax = v_mono.searchsorted( v_rolloff[1], 'right' )
        # v_mono[dexmin] >= v_rolloff[0]
        # v_mono[dexmax] > v_rolloff[1]
        nu = v_mono[dexmin:dexmax]
        spec = spec[dexmin:dexmax]    
    # If 0.25 1/cm sampling desired
    else:
        LOG.info('interpolating to 0.25cm-1')
        wn = arange(0.0,5000.0,0.25)
        dexmin = wn.searchsorted( v_mono[0], 'left' )
        dexmax = wn.searchsorted( v_mono[-1], 'right' )
        lint = interp1d(v_mono[:len(spec)], spec)
        spec = lint(wn[dexmin:dexmax])
        nu = wn[dexmin:dexmax]
    return spec,nu
    
def ifov_remap(n_snd):
    """ Map the 4 PXs within each IASI IFOV into the 2d array
    input[records][scans][detectors]{other} -> output[2*records][2*scans]{other}
    swapping 0..3 third dimension with doubling of first two dimensions
    returns a new array
    """
    dims = n_snd.shape
    redims = (dims[0]*2, dims[1]*2) + dims[3:]
    snd = empty(redims, n_snd.dtype)
    snd[0::2,1::2] = n_snd[:,:,0] # PX1
    snd[1::2,1::2] = n_snd[:,:,1] # PX2
    snd[1::2,0::2] = n_snd[:,:,2] # PX3
    snd[0::2,0::2] = n_snd[:,:,3] # PX3
    return snd

def ifov_pseudoscan(n_snd):
    """turn one scan-line variable with four detectors into two pseudo-scanlines
    input[scans,detectors,{other}] -> output[2*scans,2,{other}]
    """
    re_snd = n_snd.reshape((1,) + n_snd.shape)
    return ifov_remap( re_snd )
    
def _convert_image(detector_number):
    "array-construction functions for eugene working with or without detector selection. mildly tricky"
    if None==detector_number:
        return '',lambda A: ifov_remap(array(A))
    else:
        return '%d' % detector_number, array

def _convert_scanline(detector_number):
    if None==detector_number:
        return '',lambda A: ifov_pseudoscan(array(A))
    else:
        return '%d' % detector_number, array

RE_RECORD_INDEX = re.compile('\[(\d+)\]') # extract record number from 'mdr-1c[NNN]'

def _sounder_scanlines_all(prod, line_indices = None, only_geotemporal = False, cloud_dict = None):
    """iterate all the records in a file, returning blocks of scans as iasi_record objects.
    detector number is 0..3
    line_indices is optional parameter listing which scan lines to use, default is all
    """
    crib = None 
    scans = [int(x) for x in repeat( range(30), 2 )]  # repeat() will give us numpy.int32 by default
    # create an array of corresponding detector numbers and rearrange them the same as the input data
    dnums = [[0,1,2,3]] * 30 
    rearrange = lambda A: ifov_pseudoscan(array(A))
    wnum,dwn = sounder_wavenumbers(prod)
    detector_number = rearrange(dnums)
    if line_indices is not None:
        lines = ['mdr-1c[%d]' % x for x in line_indices]
    else:
        lines = prod._records_['mdr-1c']
    for record_name in lines: 
        GGeoSondAnglesMETOP_zen = rearrange(prod.get('%s.GGeoSondAnglesMETOP[][][0]' % (record_name)))
        GGeoSondAnglesMETOP_az = rearrange(prod.get('%s.GGeoSondAnglesMETOP[][][1]' % (record_name)))
        GGeoSondAnglesSUN_zen = rearrange(prod.get('%s.GGeoSondAnglesSUN[][][0]' % (record_name)))
        GGeoSondAnglesSUN_az = rearrange(prod.get('%s.GGeoSondAnglesSUN[][][1]' % (record_name)))
        GGeoSondLoc_lon = rearrange(prod.get('%s.GGeoSondLoc[][][0]' % (record_name)))
        GGeoSondLoc_lat = rearrange(prod.get('%s.GGeoSondLoc[][][1]' % (record_name)))        
        scanline = None if only_geotemporal else ifov_pseudoscan(array(prod.get('%s.GS1cSpect[][][]' % (record_name)),float64))
        OnboardUTC = array(prod.get( '%s.OnboardUTC' % record_name))
        CMP = {}
        if not only_geotemporal: 
            for name in CLOUD_MASK_PRODUCTS:
                CMP[name] = rearrange(prod.get( '%s.%s[][]' % (record_name,name)))
        for row in [0,1]:
            if not only_geotemporal:             
                pseudoscan = array(scanline[row])
                crib = scale_scanline(prod,pseudoscan,crib)
            R = iasi_record()
            R.wavenumbers = wnum
            R.record_index = int( *RE_RECORD_INDEX.findall(record_name) )
            R.record_name = record_name
            R.scan_number = scans   # NOTE: really is field of regard number, 0..29 
            R.detector_number = detector_number[row] # 0..3
            R.time = OnboardUTC.repeat(2)
            R.epoch = array(epoch_from_datetime(OnboardUTC)).repeat(2)
            R.radiances = None if only_geotemporal else pseudoscan
            R.longitude = GGeoSondLoc_lon[row]
            R.latitude = GGeoSondLoc_lat[row]
            R.metop_zenith_angle = GGeoSondAnglesMETOP_zen[row]
            R.metop_azimuth_angle = GGeoSondAnglesMETOP_az[row]
            R.sun_zenith_angle = GGeoSondAnglesSUN_zen[row]
            R.sun_azimuth_angle = GGeoSondAnglesSUN_az[row]
            if not only_geotemporal:
                for name in CLOUD_MASK_PRODUCTS:
                    setattr( R, name, CMP[name][row] )
            if cloud_dict:
                clod = {}
                for k in cloud_dict[(0,0,0)].keys():
                    clod[k] = array([ cloud_dict[(R.record_index,field,det)][k] for field,det in zip(R.scan_number,R.detector_number) ])
                vars(R).update(clod)    
            yield R

def _sounder_scanlines_one(prod, detector_number, line_indices = None, only_geotemporal = False, cloud_dict = None):
    """iterate all the records in a file, returning blocks of scans as iasi_record objects.
    detector number is 0..3
    line_indices is optional parameter listing which scan lines to use, default is all
    cloud_dict is an optional dictionary of {(scan#,for#,detector#) : dictionary} additional fields to merge
    """
    crib = None
    scans = range(0,30)
    wnum,dwn = sounder_wavenumbers(prod)
    if line_indices is not None:
        lines = ['mdr-1c[%d]' % x for x in line_indices]
    else:
        lines = prod._records_['mdr-1c']
    for record_name in lines:
        GGeoSondAnglesMETOP = array(prod.get('%s.GGeoSondAnglesMETOP[][%d][]' % (record_name,detector_number)))
        GGeoSondAnglesSUN = array(prod.get('%s.GGeoSondAnglesSUN[][%d][]' % (record_name,detector_number)))
        OnboardUTC = array(prod.get( '%s.OnboardUTC' % record_name))
        GGeoSondLoc = array(prod.get('%s.GGeoSondLoc[][%d][]' % (record_name,detector_number)))

        if not only_geotemporal:
            scanline = array(prod.get( '%s.GS1cSpect[][%d][]' % (record_name, detector_number)),float64)
            crib = scale_scanline(prod,scanline,crib)
        R = iasi_record()
        R.wavenumbers = wnum
        R.record_name = record_name
        R.record_index = int( *RE_RECORD_INDEX.findall(record_name) )
        R.scan_number = scans
        R.detector_number = detector_number
        R.time = OnboardUTC
        R.radiances = None if only_geotemporal else scanline
        R.longitude = GGeoSondLoc[:,0].squeeze()
        R.latitude = GGeoSondLoc[:,1].squeeze()
        R.metop_zenith_angle = GGeoSondAnglesMETOP[:,0].squeeze()
        R.metop_azimuth_angle = GGeoSondAnglesMETOP[:,1].squeeze()
        R.sun_zenith_angle = GGeoSondAnglesSUN[:,0].squeeze()
        R.sun_azimuth_angle = GGeoSondAnglesSUN[:,1].squeeze()
        if not only_geotemporal:
            for name in CLOUD_MASK_PRODUCTS:
                q = array(prod.get( '%s.%s[][%d]' % (record_name,name,detector_number)))
                setattr( R, name, q )
        if cloud_dict:
            clod = {}
            for k in cloud_dict[(0,0,0)].keys():
                clod[k] = array([ cloud_dict[(R.record_index,field,detector_number)][k] for field in R.scan_number ])
            vars(R).update(clod)
        yield R

def sounder_scanlines(prod, detector_number=None, line_indices = None, only_geotemporal = False, cloud_dict = None):
    """iterate all the records in a file, returning blocks of scans as iasi_record objects.
    detector number is 0..3
    line_indices optionally is a list of scan lines e.g. 0..150 (corresponds to the number of records in the file)
    only_geotemporal flag suppresses cloud mask and radiance data
    cloud_dict is an optional dictionary of {(scan#,for#,detector#) : dictionary} additional fields to merge
    FIXME: drop inconsistent use of "scan" in favor of "scan line", "field of regard", "field of view"
    """
    if detector_number is None: return _sounder_scanlines_all(prod,line_indices,only_geotemporal,cloud_dict=cloud_dict)
    else: return _sounder_scanlines_one(prod,detector_number,line_indices,only_geotemporal,cloud_dict=cloud_dict)
        
def sounder_records(prod, detector_number=None, line_indices = None, only_geotemporal = False, cloud_dict = None):
    """iterate all the records in a file, returning sequential scans as iasi_record objects.
    detector number is 0..3
    line_indices optionally is a list of scan lines e.g. 0..150 (corresponds to the number of records in the file)
    cloud_dict is an optional dictionary of {(scan#,for#,detector#) : dictionary} additional fields to merge
    only_geotemporal flag suppresses cloud mask and radiance data
    """
    for Q in sounder_scanlines(prod,detector_number,line_indices,only_geotemporal,cloud_dict=cloud_dict):
        R = iasi_record()
        R.record_name = Q.record_name
        R.record_index = Q.record_index
        start_scan = 0
        num_scans = len(Q.latitude)
        for s in range(start_scan, start_scan+num_scans):
            R.scan_number = Q.scan_number[s]
            R.wavenumbers = Q.wavenumbers
            R.time = [ Q.time[s] ]
            R.detector_number = Q.detector_number[s]
            R.epoch = [epoch_from_datetime(Q.time[s])]
            if not only_geotemporal:
                R.radiances = Q.radiances[s][:].squeeze()
            R.longitude = Q.longitude[s]
            R.latitude = Q.latitude[s]
            R.metop_zenith_angle = Q.metop_zenith_angle[s]
            R.metop_azimuth_angle = Q.metop_azimuth_angle[s]
            R.sun_zenith_angle = Q.sun_zenith_angle[s]
            R.sun_azimuth_angle = Q.sun_azimuth_angle[s]
            if not only_geotemporal:
                for name in CLOUD_MASK_PRODUCTS:
                    q = getattr(Q,name)
                    setattr(R,name,q[s])
            if cloud_dict:
                for k in cloud_dict[(0,0,0)].keys():
                    setattr(R,k, cloud_dict[(R.record_index,R.scan_number,R.detector_number)][k] )
            yield R

def sounder_timerange(prod):
    """return min and max time range for the data in the product
    """
    times = array(prod.get('mdr-1c[].OnboardUTC'))
    times = times.reshape(times.size)
    return min(times), max(times)

def image_location(prod,detector_number=None):
    """ return longitude, latitude arrays normalized to -180..180, -90..90
    """
    detector_number, rayray = _convert_image(detector_number)
    latlon = rayray(prod.get( 'mdr-1c[].GGeoSondLoc[][%s][]' % detector_number ))
    lon = latlon[:,:,0].squeeze()
    lat = latlon[:,:,1].squeeze()
    lat[lat<-90.] *= -1.
    lat[lat>90.] -= 180.
    lon[lon>180.] -= 360.
    lon[lon<-180.] += 360.
    return lon, lat

def sounder_image_at_wavenumber_index(prod, detector_number, wavenumber_index):
    """ return longitudes, latitudes, radiance at a given wavenumber index, given a eugene Product
    """
    lon,lat = image_location(prod,detector_number)
    detector_number, rayray = _convert_image(detector_number)
    image = rayray(prod.get('mdr-1c[].GS1cSpect[][%s][%d]' % (detector_number,wavenumber_index)))
    return lon,lat,image

def cloud_mask_images(prod,detector_number):
    """ return images of cloud information weight, #classes, Y, Z, mean, std as (name,array) pairs, given a eugene Product
    """
    out = iasi_image()
    out.longitude, out.latitude = image_location(prod,detector_number)
    detector_number, rayray = _convert_image(detector_number)
    dcloud = dict( [(s,rayray(prod.get( 'mdr-1c[].%s[][%s]' % (s,detector_number) ))) for s in CLOUD_MASK_PRODUCTS] )
    vars(out).update(dcloud)
    return out
    
def make_perimeter(lon,lat):
    """from a matrix of longitudes and latitudes, generate a simple perimeter
    """
    R,S = lon.shape
    seg1 = array(zip(lon[0:R,0],       lat[0:R,0]))
    seg2 = array(zip(lon[R-1,0:S],     lat[R-1,0:S]))
    seg3 = array(zip(lon[R-1::-1,S-1], lat[R-1::-1,S-1]))
    seg4 = array(zip(lon[0,S::-1],    lat[0,S::-1]))
    return concatenate( (seg1, seg2, seg3, seg4) ) 

def monotonize_perimeter(coord_tuples, tolerance = 90.):
    """given a perimeter as a sequence of pairs (lon,lat) 
    normalize such that perimeter coordinates are geographically monotonic
    and longitude >=-180
    and latitude >=-90
    so that we get a reasonable mechanism for doing geometric searches!
    FUTURE: promote this to a geometry_tools package
    >>> q= [ (177., 17), (178., 17), (179., 17), (-179, 18), (-178, 19), (-179, 20), (178, 21)] 
    >>> monotonize_perimeter(q)
    ((177.0, 17.0), (178.0, 17.0), (179.0, 17.0), (181.0, 18.0), (182.0, 19.0), (181.0, 20.0), (178.0, 21.0))
    """
    from numpy import subtract,zeros,cumsum
    from numpy import any as anyof
    rix = array(coord_tuples)
    dif = subtract( rix[1:,:], rix[0:-1,:] )    
    # Longitudinal fix -- 
    # find areas of big longitudinal jump
    adjust = zeros( dif.shape, dtype='d' )
    # where longitude difference jumps by -90 or more, we just went from a large -longitude to a large +longitude, subtract 360
    adjust[ dif[:,0]>tolerance ,0] -= 360.
    # where longitude difference jumps by +90 or more, we just went from a large +longitude to a large -longitude; add 360
    adjust[ dif[:,0]<-tolerance, 0] += 360.
    adjust[:,0] = cumsum(adjust[:,0])
    # FUTURE: how to deal with the poles?
    delta = concatenate( (array([(0.,0.)], dtype='d'), adjust ))
    # apply cumulative adjustments
    six = rix + delta
    if anyof( six[:,0] < -180. ): 
        six[:,0] += 360.
    return six # tuple( (lon,lat) for lon,lat in six )

def simple_metadata(prod):
    """ return dictionary of information about a product
    keys -- 
    bounding_box: ( (lon,lat), (lon,lat) ) min-max boundary
    perimeter: list of (lon,lat) as a closed polygon (perimeter[-1] == perimeter[0]) 
        i.e. 2*( (2*records) + (2*scans) ) points
    start_time, end_time: (start-time, end-time) as datetime objects
    keywords: currently blank
    state: currently blank
    description: boilerplate
    (refer to iasi_cache.py)
    FUTURE: reduce to 12 points per perimeter polygon
    """
    lon, lat = image_location(prod)
    perimeter = monotonize_perimeter(make_perimeter(lon, lat))
    plon = perimeter[:,0]
    plat = perimeter[:,1]
    box = ( (min(plon), min(plat)), (max(plon), max(plat)) )
    start,end = sounder_timerange( prod )
    return {'bounding_box': box, 
            'perimeter': perimeter, 
            'instrument': 'iasi',
            'start_time': start,
            'end_time': end,
            'keywords': ['radiance', 'image'],
            'state': '',
            'description': 'raw IASI Level 1C data'
             } 
             
def _centerpoint_nav(elon,elat,edate, olon,olat,odate):
    """ reduce the nav data to a list of 30 centerpoint longitudes, latitudes, and times
    """
    lon = [ sum(elon[n:n+2] + olon[n:n+2])/4.0 for n in range(0,60,2) ]
    lat = [ sum(elat[n:n+2] + olat[n:n+2])/4.0 for n in range(0,60,2) ]
    date = [ odate[n] for n in range(0,60,2) ]
    return zip(lon, lat), date
    
             
def generate_scanline_metadata(prod,fovs=False):
    """ metadata dictionary generator for each scan-line
    if fovs flag is set, then generates 120 fields of view instead of 30 fields of regard
    perimeter: polygon of (lon,lat) points
    start_time, end_time: (start-time, end-time) as datetime objects
    description: boilerplate
    for_center: field of regard center as a point, 30 values in a list
    for_time: field of regard time, 30 datetimes in a list
    (refer to iasi_cache.py)
    FUTURE: reduce to 12 points per perimeter polygon
    >>> import tools as iasi
    >>> for D in iasi.generate_scanline_metadata( iasi.open_product('IASI*') ): print D
    """
    gen = sounder_scanlines(prod, only_geotemporal=True)
    while True:
        try:
            # read two pseudoscan lines 60 spectra wide each
            even = gen.next()
            odd = gen.next()
            assert( even.record_index == odd.record_index )
            record_index = even.record_index
        except StopIteration: 
            break
        # there and back again simple perimeter
        perilon = concatenate( (even.longitude, odd.longitude[::-1]) )
        perilat = concatenate( (even.latitude, odd.latitude[::-1]) ) 
        perimeter = monotonize_perimeter( zip(perilon, perilat) )
        if not fovs: # then only do fields of regard - detector number is None, average 4 points for footprint center
            for_center, for_time = _centerpoint_nav( even.longitude, even.latitude, even.time, odd.longitude, odd.latitude, odd.time )
            detector_number = [None] * 30
            field_of_regard_number = range(30)
        else: # concatenate all 120 fields of view and properly label them
            for_center = zip( concatenate((even.longitude, odd.longitude)), concatenate((even.latitude, odd.latitude)) )
            for_time = concatenate( (even.time, odd.time) )
            # use list comprehensions to convert to int from numpy.int32 which psycopg2 doesn't like
            field_of_regard_number = [int(x) for x in concatenate((even.scan_number, odd.scan_number))]
            detector_number = [int(x) for x in concatenate((even.detector_number, odd.detector_number))]
        # big bucket of time points
        flattime = concatenate( (even.time, odd.time) )
        start,end = min(flattime), max(flattime)
        yield { 'perimeter': perimeter, 
                'instrument': 'iasi',
                'scan_index': record_index,
                'field_index': field_of_regard_number,   # field of regard number 0..29
                'detector_index': detector_number,  # detector number 0..3
                'start_time': start,   # scan line start time
                'end_time': end,  # scan line end time
                'for_center': for_center,   # center of a field of regard or field of view
                'for_time': for_time,  # field of view/regard time
                'description': 'raw IASI Level 1C data scan line'
                 }
        
def retrieval_read(prod, name, single_level=False):
    """pick out retrieval products from a SND format file
    reorganize geographically (120 FOVs -> 2x60FOVs, with 0,3 detectors and 1,2 detectors)
    return array of in the form value[level][0..1][0..60]
    optional scaling factor is applied as 10**(-sf)
    see pdf_ten_980760-eps-iasi-l2.pdf from EUMETSAT
    """
    prod = open_product(prod)
    lines = prod._records_['mdr']
    # read in as [scanline][ifov][level]
    sil = array( [prod.get('%s.%s' % (line,name)) for line in lines], dtype=float64 ) # * 10**(-scaling_factor)
    if single_level: 
        nlevels = 1
    else:
        nlevels = sil.shape[-1]
    nscanlines = sil.shape[0]
    # convert to [scanline][fieldofregard][detector][level] for ifov_remap use
    # ASSUMPTION: this assumes that detectors ordered 0 1 2 3 similar to the L1B products!
    # have not confirmed this in the documentation
    sfdl = reshape(sil, (nscanlines,30,4,nlevels))
    # convert to [pseudo-scanline][fieldsofview][level] using ifov_remap
    # two pseudoscanlines per scanline 
    # two fields of view per field of regard (four total)
    pvl = ifov_remap(sfdl)
    # roll level to the front
    lpv = rollaxis(pvl, 2)
    # assert(lpv is not None)
    return lpv

def retrieval_read_location(prod):
    """read EARTH_LOCATION from an SND format file
    reorganize geographically (120 FOVs -> 2x60FOVs, with 0,3 detectors and 1,2 detectors)
    return latitude[2*nscanlines][60] and longitude[same][same]
    see pdf_ten_980760-eps-iasi-l2.pdf from EUMETSAT
    """
    prod = open_product(prod)
    lines = prod._records_['mdr']
    # read in as [scanline][ifov][latlon]
    sil = array( [prod.get('%s.EARTH_LOCATION' % (line)) for line in lines], dtype=float64 )
    nscanlines = sil.shape[0]
    # convert to [scanline][fieldofregard][detector][latlon] for ifov_remap compatibility
    sfdl = reshape(sil, (nscanlines,30,4,2))
    # munge that into geographical ordering -> [pseudoscanline][fieldofview][latlon]
    pvl = ifov_remap(sfdl)
    lat = pvl[:,:,0].squeeze()
    lon = pvl[:,:,1].squeeze()
    return lat,lon

def retrieval_read_orientation(prod):
    """read ANGULAR_RELATION from an SND format file
    reorganize geographically (120 FOVs -> 2x60FOVs, with 0,3 detectors and 1,2 detectors)
    return solzen[2*nscanlines][60] and satzen[same][same], and solaz[same][same] and sataz[same][same]
    see pdf_ten_980760-eps-iasi-l2.pdf from EUMETSAT
    """
    prod = open_product(prod)
    lines = prod._records_['mdr']
    # read in as [scanline][ifov][solzensatzensolazsataz]
    sil = array( [prod.get('%s.ANGULAR_RELATION' % (line)) for line in lines], dtype=float64 )
    nscanlines = sil.shape[0]
    # convert to [scanline][fieldofregard][detector][solzensatzensolazsataz] for ifov_remap compatibility
    sfdl = reshape(sil, (nscanlines,30,4,4))
    # munge that into geographical ordering -> [pseudoscanline][fieldofview][latlon]
    pvl = ifov_remap(sfdl)
    solzen = pvl[:,:,0].squeeze()
    satzen = pvl[:,:,1].squeeze()
    solaz = pvl[:,:,2].squeeze()
    sataz = pvl[:,:,3].squeeze()
    return solzen, satzen, solaz, sataz
    
def retrieval_read_ancillary(prod):
    """return dictionary of ancillary arrays including pressure levels
    """
    anc = [ 'PRESSURE_LEVELS_TEMP', 'PRESSURE_LEVELS_HUMIDITY', 'PRESSURE_LEVELS_OZONE', 'SURFACE_EMISSIVITY_WAVELENGTHS' ]
    return dict( (name,array(prod.get('giadr.%s' % name))) for name in anc )
        

def open_product(afile):
    "given a filename, glob pattern, or eugene Product object, return a eugene Product object"
    from glob import glob
    import eugene
    if isinstance(afile,str): 
        (filename,) = glob(afile)
        return eugene.Product(filename)
    else: 
        return afile

def test_xsect_sounder(afile,detector_number,wavenumber_index):
    """show sounder wavenumber cross section
    afile: a glob pattern, filename, or eugene Product object
    detector_number: 0..3
    wavenumber_index: 0..8699
    returns dictionary of an assortment of useful things
    """
    prod = open_product(afile)
    import map_tools as mt
    lon,lat,im = sounder_image_at_wavenumber_index( prod, detector_number, wavenumber_index )
    m,x,y = mt.mapshow(lon,lat,im,resolution='c')
    return { 'latitude': lat, 'longitude': lon, 'x_transform': x, 'y_transform': y, 'basemap': m, 'image': im } 

def test_metadata(afile, do_plot=True):
    prod = open_product(afile)
    from pprint import pprint
    sm = simple_metadata(prod) 
    pprint( sm )
    if do_plot:
        poly = sm['perimeter']
        from pylab import plot, show
        (lonmin,latmin),(lonmax,latmax) = sm['bounding_box']
        quad = [ (lonmin,latmin), (lonmax,latmin), (lonmax, latmax), (lonmin,latmax), (lonmin,latmin) ]
        qx = [q[0] for q in quad]
        qy = [q[1] for q in quad]
        plot( poly[:,0], poly[:,1], 'x-', qx,qy ); show();
    return sm
    
def test_xsect_cloudmask(afile,detector_number,level_number,entity='GCcsRadAnalWgt'):
    """show cloudmask level cross section
    afile: a glob pattern, filename, or eugene Product object
    detector_number: 0..3
    level_number: 0..6
    returns dictionary of an assortment of useful things
    """
    import map_tools as mt
    prod = open_product(afile)
    lon,lat = image_location(prod,detector_number)
    D = cloud_mask_images(prod,detector_number)
    im = getattr(D,entity)[:,:,level_number]
    m,x,y = mt.mapshow(D.longitude,D.latitude,im,resolution='c')
    vars(D).update( { 'x_transform': x, 'y_transform': y, 'basemap': m, 'image': im } )
    return D
    
def test_xsect_radanalmean(afile,detector,level1,level2):
    """given filename pattern or product object, detector number (0..3), level(0..7) and species? (0..5)
    plot out a contour map on an image
    """
    import map_tools as mt
    prod = open_product(afile)
    D = cloud_mask_images(prod,detector)
    im = D.GCcsRadAnalMean[:,:,level1,level2]
    m,x,y = mt.mapshow(D.longitude, D.latitude, im,resolution='c' )
    vars(D).update( { 'x_transform': x, 'y_transform': y, 'basemap': m, 'image': im } )
    return D

