#!/usr/bin/env python
# encoding: utf-8
"""
iasi_cache.py

Created by Keoni on 2007-11-29.
Copyright (c) 2007 University of Wisconsin SSEC. All rights reserved.

Setting up a cache:
Create database
initialize database
create IASI_cache directory on SAN
get upstream directory set up

ssh -L5432:bora2:5432 ash
python -m iasi_cache --iasi-db=postgres://rayg:c@nnab1s@localhost/iasi_cache 

requires miscellaneous routines from keoni 0.3.4 or better (see larch.ssec.wisc.edu)

"""

_CVS_IDENT = "$Id: iasi_cache.py,v 1.124 2008/08/18 21:38:21 rayg Exp $"

import sys
import os
import logging
import uuid
import re
import time
from cPickle import dump
from ftplib import FTP
from subprocess import check_call as sh
from datetime import datetime, timedelta
from time import gmtime, strftime
from collections import defaultdict

from sqlalchemy import *    
import sqlalchemy.orm as orm

LOG = logging.getLogger( __name__ )

from metahoard.iasi import tools as iasi
from metahoard.iasi.config import *
from metahoard.iasi.schema import *
from metahoard.iasi.util import *


# identifying a shopping list file from a list of FTP files, and returning the zone of interest as \1
RE_SHOPPING_LIST = re.compile('cgetiasi\.moa\.(?:\d+)(?:\.(.*))?')

# identifying an IASI raw data file
RE_IASI = re.compile('IASI_xxx_1C_[_A-Z0-9]+')

# def _congeal(*tuptup):
#     from operations
#     if len(tuptup)==1: return tuptup[0]
#     return ''.join(congeal(x) for x in tuptup)
    
TMP = os.path.join( os.environ.get('TMP','/tmp'), 'iasi_cache.%d' % os.getpid() )

def _get_available_files((site,wd) = UPSTREAM_FTP):
    """look upstream and get a set of filenames which are available
    """
    ftp = FTP(site)
    ftp.login()
    ftp.cwd(wd)
    return set([fn for fn in ftp.nlst() if '.xml' not in fn and "IASI" in fn])

def _download_file((site,wd), filename):
    "download a file and return local pathname"
    ftp = FTP(site)
    ftp.login()
    ftp.cwd(wd)
    local_pn = os.path.join(TMP,filename)
    fp = file(local_pn,'wb')
    def stor(data,fp=fp):
        fp.write(data)
    LOG.info('downloading %s to %s' % (filename, local_pn))
    ftp.retrbinary('RETR %s' % filename, stor)
    fp.close()
    ftp.close()
    return local_pn

def _upload_to_cache(local_pathname):
    """send a file to the cache, returning a list of urls which it can be accessed at
    """
    LOG.debug('uploading %s to cache' % local_pathname)
    sh( ['rsync', '-au', '--progress', local_pathname, CACHE_RSYNC_PUT] )
    return [ CACHE_RSYNC + os.path.split(local_pathname)[-1] ]

def _upload_mask_to_storage(name,data):
    "given the contenst of an iasimask file, and the iasi raw data filename associated with it, send the mask to the PEATE"
    fn = os.path.join(TMP,'iasimask_%s.txt' % name)
    file(fn,'wt').write(data)
    LOG.info('delivering iasimask file to long-term storage')
    sh(['rsync', '-au', fn, CACHE_RSYNC_PUT])
    os.unlink(fn)

def _download_file_as_string(ftp, filename):
    """ return the content of an FTP RETR
    """
    buffr = []
    def stor(data, s=buffr):
        LOG.debug("received %d bytes" % len(data))
        s.append(data)
    ftp.retrbinary('RETR %s' % filename,stor)
    return ''.join(buffr)

def _monkey_ftp_session( now, (site,wd) ):
    """ given a datetime object and a working directory, return an FTP session
    """
    ftp = FTP(site,user='ftp',passwd='iasi_cache@ssec.wisc.edu')
    if not now: now = datetime.utcnow() - timedelta(1) # yesterday
    wd = now.strftime(wd)
    ftp.cwd(wd)
    return ftp

def read_shopping_lists( now = None ):
    """ for a given time tuple (defaults to now) 
        generate (iasi-filename, zone) tuples
        see information on katamari.ssec.wisc.edu from HalW
    """
    for (site,dirname,kword) in SHOPPING_LIST:        
        ftp = _monkey_ftp_session(now, (site,dirname))
        fnlist = ftp.nlst()
        LOG.debug( 'files available: %s' % fnlist )
        for fn in fnlist:
            m = RE_SHOPPING_LIST.match(fn)
            if not m: continue
            zone = m.groups() # e.g. australia
            if kword is not None: zone=kword
            else: zone = zone[0]
            LOG.info('reading %s in zone %s' % (fn,zone))
            # read the file into a string
            content = _download_file_as_string(ftp,fn)
            LOG.debug( 'read %d bytes from FTP link' % len(content) )
            # get all the filenames in the text and yield tuples
            allfiles = RE_IASI.findall(content) 
            LOG.debug( 'filenames found in shopping list: %s' % allfiles )
            for iasifn in allfiles:
                yield iasifn, zone

def shoplist_to_keywords(gen):
    """ given the output of read_shopping_lists, return a dictionary of { filename: [zone1, zone2, ..] }
    """
    ret = { }
    for fn,zone in gen:
        D = ret.setdefault(fn,[])
        D.append(zone)
    return ret
    
