#!/usr/bin/env python
# encoding: utf-8
"""
iasi_cache.py

Created by Keoni on 2007-11-29.
Copyright (c) 2007 University of Wisconsin SSEC. All rights reserved.

Setting up a cache:
Create database
initialize database
create IASI_cache directory on SAN
get upstream directory set up

ssh -L5432:bora2:5432 ash
python -m iasi_cache --iasi-db=postgres://rayg:c@nnab1s@localhost/iasi_cache 

requires miscellaneous routines from keoni 0.3.4 or better (see larch.ssec.wisc.edu)

"""

_CVS_IDENT = "$Id: iasi_cache.py,v 1.124 2008/08/18 21:38:21 rayg Exp $"

import sys
import os
import logging
import uuid
import re
import time
from cPickle import dump
from ftplib import FTP
from subprocess import check_call as sh
from datetime import datetime, timedelta
from time import gmtime, strftime
from collections import defaultdict

from sqlalchemy import *    
import sqlalchemy.orm as orm

from metahoard.iasi import tools as iasi
from metahoard.iasi.config import *
from metahoard.iasi.schema import *
from metahoard.iasi.transfer import *

LOG = logging.getLogger( __name__ )

#
# High level functions
#

# RE_DATE = re.compile( r'(\d{4})(\d{2})(\d{2}):(\d{2})(\d{2})(\d{2})'

def peate_urlget(year,day,lookfor=None):
    """given julian day, return list of RSYNC addresses for nat data for IASI
    e.g. peate_urlget(2007,100)
    """
    from re import findall
    from urllib2 import urlopen
    julday = '%04d%03d' % (year,day)
    txt = (urlopen('http://bora7.ssec.wisc.edu/query/show_day.py?typ=15&jdate=' + julday).read() + 
           urlopen('http://bora7.ssec.wisc.edu/query/show_day.py?typ=10&jdate=' + julday).read())
    zult = [ 'rsync://%s/%s' % q for q in findall( '(\w+)::(\S+)', txt ) ]
    return [x for x in zult if lookfor in x] if lookfor else zult

def peate_urimake(name):
    """given the name of an iasi granule, return a PEATE uri at which it should be found
    """
    # extract YYYYMMDD
    from re import findall
    from time import mktime,gmtime
    (ymdstr,) = findall('IASI_xxx_1C_M02_(\d{8})',name)
    yyyy = int(ymdstr[:4])
    mm = int(ymdstr[4:6])
    dd = int(ymdstr[6:8])
    t = mktime((yyyy, mm, dd, 0, 0, 0, 0, 0, 0))
    # compute julian day
    jday = gmtime(t)[7]
    return PYTHON_URI_BASE + "peate_urlget(%04d,%02d,'%s')" % (yyyy,jday,name)

def ui_query(db=None, box=None, startend=None, keywords=None, for_list=None, string_criteria=None, yield_objects=False):
    """ command line query handler: parse query, recover information from database and yield field of regard info
    given a bounding box string (lon,lat),(lon,lat) or compatible with keoni.map.parse:parse_coords
    and YYYYMMDD:HHMMSS-YYYYMMDD:HHMMSS for time range (keoni.time.epoch:datetime_from_string)
    generate dictionaries of information about the resource to download and which scans/fields to use
        name: filename
        urls: relevant replica download URLs
        keywords: dictionary of key-value pairs
        fields: { scanline-index: { field-index: set(detector_indices), ... }, ... }
    """
    if not db: 
        db = CacheDatabase(DATABASE_URL)
        
    if startend is not None and startend is not '-':
        startend = [ datetime_from_string(x.strip()) for x in startend.split('-') ]
    else:
        startend = None
        
    if box is not None and box is not '-':
        if box in BOUNDING_BOXES: 
            # print "using bounding box %s" % '-'.join([format_coords(x) for x in box])
            box = BOUNDING_BOXES[box]
            print str(box)
        else: # try using parse_coords, fall back to (lon,lat) pairs
            if '(' in box and ',' in box and ')' in box: 
                LOG.info('using old lon,lat pairs format')
                box = eval( '(%s)' % box )
            else:    
                box = tuple(parse_coords(box))
        LOG.info("search box is %s" % ', '.join(format_coords(x) for x in box))
    else: # skip bounding box constraint
        box = None
        
    if keywords is not None and keywords is not '-':
        # convert to a pythonic keyword expression - FUTURE this doesn't work anywhere near right
        # keywords = ' and '.join( 'bool(%s)' % s.strip() for s in keywords.split(',') )
        keywords = set([x.strip() for x in keywords.split(',')])
        # print keywords
    else: 
        keywords = None
        
    # refine by fields of regard
    if for_list is not None and for_list is not '-':
        # subs = dict(ALL=range(30),CENTER=[14,15])
        for_list = eval( '['+for_list+']' )
        LOG.info('field of regards limited to %s' % ', '.join(str(x) for x in for_list))
    else:
        for_list=None
        
    # we need a least a bounding box or a time domain to be able to constrain things sufficiently
    # (for now)
    if not box and not startend:
        print "no bounding box or bounding time provided! exiting.."
        return
    
    # FIXME: make the keyword constraints part of search_fields_*
    
    if yield_objects:
        for field_obj in db.search_fields_as_objects(time_range = startend, bounding_box = box, for_list=for_list, string_criteria=string_criteria): 
            yield field_obj
        return
    
    # create a search generator
    zult = db.search_fields_of_regard(time_range = startend, bounding_box = box, for_list=for_list, string_criteria=string_criteria) # if box else db.search_scans( time_range = startend )
    for uu,fields in zult:
        name = db.name_for_uuid(uu)
        urls = db.urls_for_uuid(uu)
        kws = db.keywords_for_uuid(uu)
        if keywords: # filter by keywords by hand for now - FUTURE: do this server side
            K = set(kws.keys())
            if K.intersection(keywords) != keywords: 
                LOG.debug('skipping %s due to keywords not being a subset' % name)
                continue
        yield { 'name': name, 
                'urls': urls, 
                'keywords': kws, 
                'fields': fields
                }

_ui_query = ui_query

def _keyword_string(keywords):
    "return a string explaining keywords"
    return ', '.join(("%s=%s" % (k,v)) if v is not None else ('%s' % k) for k,v in keywords.items())

def _format_search_results(scanfors):
    zult = []
    for scanline,fields in scanfors.items():
        zult.append( 'line %d: ' % scanline + ' '.join( '%d.%d' % (f,d) for f in fields for d in fields[f] ) )
    return '\n'.join(zult) + '\n'

def query(db=None, box=None, startend=None, keywords=None, fields=None, string_criteria=None):
    """ given string in the format (lon,lat),(lon,lat) for bounding box
    and YYYYMMDD:HHMMSS-YYYYMMDD:HHMMSS for time range
    print out a list of names, scans, and records 
    with associated URLs to download from
    """
    for D in ui_query(db,box,startend,keywords,fields,string_criteria):
        name = D['name']
        urls = D['urls']
        keywords = D['keywords']
        fields = D['fields']
        print '-'*72
        print name
        if keywords: 
            print "keywords: " + _keyword_string(keywords)
        print '\n'.join(urls)
        print _format_search_results(fields)

def sort_urls_by_score(urls):
    """quick heuristic set for deciding which url is going to be the best to use.
    stupid hard-codey stuff in here.
    returns list of urls, sorted by preference
    """
    scor = dict( (url,0) for url in urls )
    for k in scor: 
        # 'rsync://sands' get a +4 score
        if 'rsync://sands' in k:
            scor[k] += 4
        # 'ftp' links get a +1 score
        if 'ftp://' in k:
            scor[k] += 1
        if PYTHON_URI_BASE in k: 
            scor[k] += 2
    # sort them in scor order and grab the last one
    tally = [ (v,k) for k,v in scor.items() ]
    tally.sort()
    tally.reverse()
    return [x[1] for x in tally]
    
def emit_detectors_from_search(pathname,scandict):
    """for each scan of interest in an IASI native file
    provide individual spectra appropriate to store as files of the form iasi_YYYYMMDD_HHMMSS_sMM_fNN_dD.mat
    MM = scan number, NN is field number, D is detector number
    """
    prod = iasi.open_product(pathname)
    all_scans = list(scandict.keys())
    gen = iasi.sounder_scanlines( prod, line_indices = all_scans )
    while True: 
        try:
            # read two pseudoscan lines 60 spectra wide each
            even = gen.next()
            odd = gen.next()
            assert( even.record_index == odd.record_index )
            record_index = even.record_index
        except:
            break
        # get list of fields-of-regard of interest
        fields = scandict[record_index]
        for field,detectors in fields.items():
            # there are four radiance spectra per field-of-regard 
            # two at even.radiances[field*2], two in odd
            # likewise with .time and all that
            T = even.time[field*2].strftime('%Y%m%d_%H%M%S')
            M = record_index
            N = field
            R = { }
            R['epoch'] = even.epoch[field*2]
            R['wavenumbers'] = even.wavenumbers
            nwn = len(even.wavenumbers)
            R['scan_index'] = record_index
            R['creator'] = _CVS_IDENT
            R['source_file'] = os.path.split(pathname)[1]
            # now, yield each of the four detectors in turn
            abba = [(line,index) for line in (even,odd) for index in (field*2,field*2+1)]
            for line,dex in abba: 
                D = line.detector_number[dex]
                if D not in detectors: 
                    LOG.debug('skipping detector %d, not in %s' % (D,detectors))
                    continue
                R['field_index'] = field
                R['detector_index'] = D
                R['filename'] = 'iasi_%(T)s_s%(M)02d_f%(N)02d_d%(D)d.mat' % locals()
                R['radiances'] = line.radiances[dex][:nwn]
                R['longitude'] = line.longitude[dex]
                R['latitude'] = line.latitude[dex]
                R['metop_zenith_angle'] = line.metop_zenith_angle[dex]
                R['metop_azimuth_angle'] = line.metop_azimuth_angle[dex]
                R['sun_zenith_angle'] = line.sun_zenith_angle[dex]
                R['sun_azimuth_angle'] = line.sun_azimuth_angle[dex]
                yield R
                
PYTHON_URI_BASE = 'python:iasi_cache#'
def _get(remote,local,may_decompress=True):
    """ take a replica url and turn it into a local file
    if replica url is of the form python:expression(parameters), evaluate it to get a list of urls
    then try each url until one works
    decompress .gz and .bz2 data
    FUTURE: move this into keoni.net.urllib.get() with a decompress=True default parameters
    """
    from keoni.net.urllib import get 
    if PYTHON_URI_BASE == remote[0:len(PYTHON_URI_BASE)]:# then iam an expression
        urls = eval(remote[len(PYTHON_URI_BASE):])  # RUDE! FIXME: constrain to a limited symbol table
        LOG.info( 'generated urls from python URI: %s' % urls )
    else:
        urls = [remote]
    for url in urls: 
            LOG.debug('downloading "%s" -> %s' % (url,local))
        # try:
            # decompress compressed files from the peate
            if '.bz2' == url[-4:] and may_decompress:
                LOG.debug('getting bzipped file')
                get(url, local + '.bz2')
                LOG.debug('bunzipping')
                sh(['bunzip2',local + '.bz2'])
            elif '.gz' == url[-3:] and may_decompress:
                LOG.debug('getting gzipped file')
                get(url, local + '.gz')
                LOG.debug("gunzipping")
                sh(['gunzip',local + '.gz'])                
            else:
                LOG.debug('getting simple file')
                get(url,local)
            break
        # except:
        #     pass
        
def get_any_replica(remotes,local_pathname,force=False):
    """same as _get, but allows multiple remote URLs
    returns chosen URL, or None if download failed
    returns local pathname if file is already available, except when forced to download
    FIXME: ftp urls don't work???
    """
    chosen_url = None
    if os.path.exists(local_pathname) and not force:
        LOG.info('file %s is already available, skipping download' % local_pathname)
        return local_pathname
    for url in remotes: 
        LOG.debug('attempting to download %s to %s' % (url,local_pathname))
        try:
            _get(url,local_pathname)
            assert(os.path.exists(local_pathname))
            # FUTURE: if this fails then remove that url from the replica table, possibly with a callable parameter to handle
            chosen_url = url
            LOG.info('downloaded %s as %s' % (url,local_pathname))
            break
        except:
            LOG.debug('failed to download %s!' % url)
    return chosen_url
                                       
def obtain_granule(db, target_dir, granule_name):
    """return granule name, local pathname, chosen url, list of scanline indices
    input is a granule name to be downloaded, or a filename of a granule to use
    as well as a temporary directory to situate any downloaded file into
    local pathname is None if the file was unavailable
    chosen url is None if the file was available locally
    """
    # is the file already local?
    if os.path.exists(granule_name):
        local_pathname = granule_name
        granule_name = os.path.split(local_pathname)[1]
        chosen_url = local_pathname
        LOG.info('using local file '+local_pathname)
        linenums = range( len( iasi.open_product(local_pathname)._records_['mdr-1c'] ) )
        chosen_url = None
    else: # consult the database and download it
        S = db.session()
        (g,) = S.query(Granule).filter_by(name=granule_name).all()
        replicas = db.urls_for_uuid(g.uuid)
        urls = _sort_urls_by_score(replicas)
        LOG.debug('checking urls %s' % urls)
        local_pathname = os.path.join(target_dir, granule_name)
        chosen_url = _get_any_replica(urls,local_pathname)
        if not chosen_url: local_pathname = None
        linenums = [s.scan_index for s in g.scans]
    return granule_name, local_pathname, chosen_url, linenums

def emit_mat_files_from_search(target_dir, url, local_pathname, fields):
    import scipy.io.mio as mio
    for R in _emit_detectors_from_search(local_pathname,fields):
        tgt_name = R['filename']
        print 'writing %s...' % tgt_name
        tgt_path = os.path.join(target_dir, tgt_name)
        R['source_url'] = url
        mio.savemat( tgt_path, R )
        
def expand_user_field_selection(linefields,linenums):
    """return a revised scan line and field selection dictionary, expanding scan line tuples 
    see help for 'export' for details on the dictionary being expanded
    """
    # make a copy
    linefields = dict(linefields)
    # check if ALL was included and expand it
    if None in linefields:
        allfields = linefields[None]
        # if anything was explicitly mentioned, treat ALL as a default
        morefields = dict( (num,allfields) for num in linenums if num not in linefields )
        del linefields[None]
        linefields.update(morefields)
    # check for lists in the keys and expand those
    # also expand all-fields
    finalfields = {}
    allfovs = set([0,1,2,3])
    for k,v in linefields.items():  # k is scanline number, v is field list
        if v is None: 
            v = dict( (x,allfovs) for x in range(30) )
        else:
            if isinstance(v,int):
                v = { v: allfovs }
            else:
                v = dict( (x,allfovs) for x in v )
        if isinstance(k,tuple):
            for subk in k:
                finalfields[subk] = v
        else:
            finalfields[k] = v
    return finalfields
